package net.sf.jpatch.auxilary;

import javax.vecmath.*;

/**
 * This class provides static methods to tasselate bicubic Bezier patches
 */
public final class Tesselator {
	private static double[][] aaB0;
	private static double[][] aaB1;
	private static double[][] aaB2;
	private static double[][] aaB3;
	private static double[][] aadB0;
	private static double[][] aadB1;
	private static double[][] aadB2;
	private static double[][] aadB3;

	private static Point3d[][][] aaap3P;
	private static Vector3d[][][] aaav3N;
	private static Point3d[] p = new Point3d[16];

	private static Point3d[][] points;
	private static Vector3d[][] normals;
	
	static {
		init(66);
	}

	/**
	 * precompute Bernstein coefficients
	 */
	public static void init(int maxSubdiv) {
		aaB0 = new double[maxSubdiv][];
		aaB1 = new double[maxSubdiv][];
		aaB2 = new double[maxSubdiv][];
		aaB3 = new double[maxSubdiv][];
		aadB0 = new double[maxSubdiv][];
		aadB1 = new double[maxSubdiv][];
		aadB2 = new double[maxSubdiv][];
		aadB3 = new double[maxSubdiv][];
		aaap3P = new Point3d[maxSubdiv][][];
		aaav3N = new Vector3d[maxSubdiv][][];

		for (int sd = 2; sd < maxSubdiv; sd++) {
			aaB0[sd] = new double[sd];
			aaB1[sd] = new double[sd];
			aaB2[sd] = new double[sd];
			aaB3[sd] = new double[sd];
			aadB0[sd] = new double[sd];
			aadB1[sd] = new double[sd];
			aadB2[sd] = new double[sd];
			aadB3[sd] = new double[sd];
			aaap3P[sd] = new Point3d[sd][sd];
			aaav3N[sd] = new Vector3d[sd][sd];
			for (int i = 0; i < sd; i++) {
				double s = (double) i / (sd - 1);
				double s1 = 1 - s;
				double s2 = s * s;
				double s12 = s1 * s1;
				aaB0[sd][i] = s12 * s1;
				aaB1[sd][i] = 3 * s * s12;
				aaB2[sd][i] = 3 * s2 * s1;
				aaB3[sd][i] = s2 * s;
				aadB0[sd][i] = - 3 * s12;
				aadB1[sd][i] = -6 * s * s1 + 3 * s12;
				aadB2[sd][i] = -3 * s2 + 6 * s * s1;
				aadB3[sd][i] = 3 * s2;
				for (int j = 0; j < sd; j++) {
					aaap3P[sd][i][j] = new Point3d();
					aaav3N[sd][i][j] = new Vector3d();
				}
			}
		}
		for (int i = 0; i < 16; i++) {
			p[i] = new Point3d();
		}
	}

	public static Point3d[][] getPointArray() {
		return points;
	}
	
	public static Vector3d[][] getNormalArray() {
		return normals;
	}
	
	/**
	 * tesselate a Bezier patch
	 */
	public static void tesselate(Point3f[] pf, int subdiv) {
		double x0, x1, x2, x3, y0, y1, y2, y3, z0, z1, z2, z3;
		double[] B0, B1, B2, B3, dB0, dB1, dB2, dB3;
		Vector3d v3ds = new Vector3d();
		Vector3d v3dt = new Vector3d();
		Vector3d v3n = new Vector3d();

		for (int i = 0; i < 16; i++) {
			p[i].set(pf[i]);
		}

		if (subdiv < aaB0.length) {

			/* use precomputed Bernstein coefficients */
			points = aaap3P[subdiv];
			normals = aaav3N[subdiv];
			B0 = aaB0[subdiv];
			B1 = aaB1[subdiv];
			B2 = aaB2[subdiv];
			B3 = aaB3[subdiv];
			dB0 = aadB0[subdiv];
			dB1 = aadB1[subdiv];
			dB2 = aadB2[subdiv];
			dB3 = aadB3[subdiv];
		} else {

			/* compute new Bernstein coefficients */
			points = new Point3d[subdiv][subdiv];
			normals = new Vector3d[subdiv][subdiv];
			B0 = new double[subdiv];
			B1 = new double[subdiv];
			B2 = new double[subdiv];
			B3 = new double[subdiv];
			dB0 = new double[subdiv];
			dB1 = new double[subdiv];
			dB2 = new double[subdiv];
			dB3 = new double[subdiv];
			for (int i = 0; i < subdiv; i++) {
				double s = (double) i / (subdiv - 1);
				double s1 = 1 - s;
				double s2 = s * s;
				double s12 = s1 * s1;
				B0[i] = s12 * s1;
				B1[i] = 3 * s * s12;
				B2[i] = 3 * s2 * s1;
				B3[i] = s2 * s;
				dB0[i] = - 3 * s12;
				dB1[i] = -6 * s * s1 + 3 * s12;
				dB2[i] = -3 * s2 + 6 * s * s1;
				dB3[i] = 3 * s2;
				for (int j = 0; j < subdiv; j++) {
					points[i][j] = new Point3d();
					normals[i][j] = new Vector3d();
				}
			}
		}

		/* compute points and normals */
		for (int s = 0; s < subdiv; s++) {
			for (int t = 0; t < subdiv; t++) {
				x0 = p[0].x * B0[t] + p[1].x * B1[t] + p[2].x * B2[t] + p[3].x * B3[t];
				x1 = p[4].x * B0[t] + p[5].x * B1[t] + p[6].x * B2[t] + p[7].x * B3[t];
				x2 = p[8].x * B0[t] + p[9].x * B1[t] + p[10].x * B2[t] + p[11].x * B3[t];
				x3 = p[12].x * B0[t] + p[13].x * B1[t] + p[14].x * B2[t] + p[15].x * B3[t];
				y0 = p[0].y * B0[t] + p[1].y * B1[t] + p[2].y * B2[t] + p[3].y * B3[t];
				y1 = p[4].y * B0[t] + p[5].y * B1[t] + p[6].y * B2[t] + p[7].y * B3[t];
				y2 = p[8].y * B0[t] + p[9].y * B1[t] + p[10].y * B2[t] + p[11].y * B3[t];
				y3 = p[12].y * B0[t] + p[13].y * B1[t] + p[14].y * B2[t] + p[15].y * B3[t];
				z0 = p[0].z * B0[t] + p[1].z * B1[t] + p[2].z * B2[t] + p[3].z * B3[t];
				z1 = p[4].z * B0[t] + p[5].z * B1[t] + p[6].z * B2[t] + p[7].z * B3[t];
				z2 = p[8].z * B0[t] + p[9].z * B1[t] + p[10].z * B2[t] + p[11].z * B3[t];
				z3 = p[12].z * B0[t] + p[13].z * B1[t] + p[14].z * B2[t] + p[15].z * B3[t];
				points[s][t].set(
					x0 * B0[s] + x1 * B1[s] + x2 * B2[s] + x3 * B3[s],
					y0 * B0[s] + y1 * B1[s] + y2 * B2[s] + y3 * B3[s],
					z0 * B0[s] + z1 * B1[s] + z2 * B2[s] + z3 * B3[s]
				);
				v3ds.set(
					x0 * dB0[s] + x1 * dB1[s] + x2 * dB2[s] + x3 * dB3[s],
					y0 * dB0[s] + y1 * dB1[s] + y2 * dB2[s] + y3 * dB3[s],
					z0 * dB0[s] + z1 * dB1[s] + z2 * dB2[s] + z3 * dB3[s]
				);
				v3dt.set(
					B0[s] * (p[0].x * dB0[t] + p[1].x * dB1[t] + p[2].x * dB2[t] + p[3].x * dB3[t]) +
						B1[s] * (p[4].x * dB0[t] + p[5].x * dB1[t] + p[6].x * dB2[t] + p[7].x * dB3[t]) +
						B2[s] * (p[8].x * dB0[t] + p[9].x * dB1[t] + p[10].x * dB2[t] + p[11].x * dB3[t]) +
						B3[s] * (p[12].x * dB0[t] + p[13].x * dB1[t] + p[14].x * dB2[t] + p[15].x * dB3[t]),
					B0[s] * (p[0].y * dB0[t] + p[1].y * dB1[t] + p[2].y * dB2[t] + p[3].y * dB3[t]) +
						B1[s] * (p[4].y * dB0[t] + p[5].y * dB1[t] + p[6].y * dB2[t] + p[7].y * dB3[t]) +
						B2[s] * (p[8].y * dB0[t] + p[9].y * dB1[t] + p[10].y * dB2[t] + p[11].y * dB3[t]) +
						B3[s] * (p[12].y * dB0[t] + p[13].y * dB1[t] + p[14].y * dB2[t] + p[15].y * dB3[t]),
					B0[s] * (p[0].z * dB0[t] + p[1].z * dB1[t] + p[2].z * dB2[t] + p[3].z * dB3[t]) +
						B1[s] * (p[4].z * dB0[t] + p[5].z * dB1[t] + p[6].z * dB2[t] + p[7].z * dB3[t]) +
						B2[s] * (p[8].z * dB0[t] + p[9].z * dB1[t] + p[10].z * dB2[t] + p[11].z * dB3[t]) +
						B3[s] * (p[12].z * dB0[t] + p[13].z * dB1[t] + p[14].z * dB2[t] + p[15].z * dB3[t])
				);
				v3n.cross(v3ds, v3dt);
				v3n.normalize();
				normals[s][t].set(v3n);
			}
		}
	}
}
