package net.sf.jpatch.control.edit;

import java.util.*;
import net.sf.jpatch.auxilary.*;
import net.sf.jpatch.entity.*;
import net.sf.jpatch.boundary.*;
import net.sf.jpatch.boundary.selection.*;

/**
* This abstract class provides common methods to clone ControlPoints which are
* needed by CopyEdit, LatheEdit and ExtrudeEdit
**/

public abstract class CloneCommonEdit extends JPatchCompoundEdit {

	/** Array containing all ControlPoints to be cloned (in most cases the selection) **/
	protected ControlPoint[] acp;
	/** maps cloned ControlPoints to their originals **/
	protected Map mapClones = new HashMap();
	/** maps originals to their clones **/
	protected Map mapOriginals = new HashMap();

	public CloneCommonEdit(ControlPoint[] controlPointsToClone) {
		acp = controlPointsToClone;
	}

	/**
	* checks all selected ControlPoints (those in acp), and if they are really to be
	* cloned, adds them to mapCPs
	**/
	protected void buildCloneMap(boolean hooks) {
		for (int i = 0; i < acp.length; i++) {
			ControlPoint[] acpStack = acp[i].getStack();
			for (int s = 0; s < acpStack.length; s++) {
				ControlPoint cp = acpStack[s];

				/*
				* check if a hook-curve should be added
				*/
				if (cp.isStartHook()) {
					if (hooks && checkStartHook(cp)) {
						addToCloneMap(cp);
						addToCloneMap(cp.getEnd());
					}
				} else if (cp.isHook()) {
					if (hooks && cp.getPrevAttached() != null && checkForNeighbor(cp.getPrevAttached())) {
						if (checkStartHook(cp.getStart())) {
							addToCloneMap(cp);
						}
					}
				} else if (checkForNeighbor(cp)) {
					addToCloneMap(cp);
				}
			}
		}
	}

	/**
	* configures the clones after their originals
	**/
	protected void cloneControlPoints() {
		for (Iterator it = mapOriginals.keySet().iterator(); it.hasNext(); ) {
			ControlPoint cpClone = (ControlPoint) it.next();
			ControlPoint cpOriginal = getOriginal(cpClone);

			/* set next and prev controlPoints */
			if (!cpOriginal.isHook()) {
				cpClone.setNext(getClone(cpOriginal.getNext()));
				cpClone.setPrev(getClone(cpOriginal.getPrev()));
			} else {
				ControlPoint cpNext;
				for (cpNext = cpOriginal.getNext(); cpNext != null && getClone(cpNext) == null; cpNext = cpNext.getNext());
				cpClone.setNext(getClone(cpNext));
				ControlPoint cpPrev;
				for (cpPrev = cpOriginal.getPrev(); cpPrev != null && getClone(cpPrev) == null; cpPrev = cpPrev.getPrev());
				cpClone.setPrev(getClone(cpPrev));
			}
			
			/* set next attached and prev attached controlPoints */
			ControlPoint cpNextAttached;
			for (cpNextAttached = cpOriginal.getNextAttached(); cpNextAttached != null && getClone(cpNextAttached) == null; cpNextAttached = cpNextAttached.getNextAttached());
			cpClone.setNextAttached(getClone(cpNextAttached));
			ControlPoint cpPrevAttached;
			for (cpPrevAttached = cpOriginal.getPrevAttached(); cpPrevAttached != null && getClone(cpPrevAttached) == null; cpPrevAttached = cpPrevAttached.getPrevAttached());
			cpClone.setPrevAttached(getClone(cpPrevAttached));

			/* set parent and child hooks */
			cpClone.setParentHook(getClone(cpOriginal.getParentHook()));
			cpClone.setChildHook(getClone(cpOriginal.getChildHook()));
			
			/* set hook pos */
			cpClone.setHookPos(cpOriginal.getHookPos());
		}
	}

	/**
	* add cloned curves to model
	**/
	protected void cloneCurves() {
		for (Iterator it = mapOriginals.keySet().iterator(); it.hasNext(); ) {
			ControlPoint cpClone = (ControlPoint) it.next();
			ControlPoint cpOriginal = getOriginal(cpClone);

			/* check for a loop */
			if (cpOriginal.getLoop()) {
				boolean bLoop = false;
				loop:
				for (ControlPoint cp = cpClone.getNext(); cp != null; cp = cp.getNext()) {
					if (cp == cpClone) {
						bLoop = true;
						break loop;
					}
				}
				cpClone.setLoop(bLoop);
			}

			/* if we are the start, add the curve */
			if (cpClone.getLoop() || cpClone.getPrev() == null) {
				if (cpClone.getNext() != null) {
					Curve curve = new Curve(cpClone);
					addEdit(new CreateCurveEdit(curve));
					addEdit(new ValidateCurveEdit(curve));
				} else {
					System.err.println("Error in CloneCommonEdit - attempted to add invalid curve");
				}
			}
		}
	}

	/**
	* create new selection
	**/
	protected PointSelection createNewSelection() {
		PointSelection ps = new PointSelection();
		for (Iterator it = mapOriginals.keySet().iterator(); it.hasNext(); ) {
			ControlPoint cpClone = (ControlPoint) it.next();
			if (cpClone.isHead()) {
				ps.addControlPoint(cpClone);
			}
		}
		return ps;
	}

	/**
	* clone patches
	**/
	protected void clonePatches() {
		List list = new ArrayList();
		Model model = MainFrame.getInstance().getModel();
		for (Patch patch = model.getFirstPatch(); patch != null; patch = patch.getNext()) {
			ControlPoint[] acpOriginalPatch = patch.getControlPoints();
			ControlPoint[] acpClonePatch = new ControlPoint[acpOriginalPatch.length];
			boolean addPatch = true;
			loop:
			for (int n = 0; n < acpOriginalPatch.length; n++) {
				if (getClone(acpOriginalPatch[n]) == null) {
					addPatch = false;
					break loop;
				} else {
					acpClonePatch[n] = getClone(acpOriginalPatch[n]);
				}
			}
			if (addPatch) {
				list.add(acpClonePatch);
				list.add(patch.getMaterial());
				if (acpClonePatch.length == 10) {
					ControlPoint[] acp5 = new ControlPoint[] {
						acpClonePatch[0].trueHead(),
						acpClonePatch[2].trueHead(),
						acpClonePatch[4].trueHead(),
						acpClonePatch[6].trueHead(),
						acpClonePatch[8].trueHead()
						};
					model.getCandidateFivePointPatchList().add(acp5);
				}
			}
		}
		for (Iterator it = list.iterator(); it.hasNext(); ) {
			Patch patch = new Patch((ControlPoint[]) it.next());
			patch.setMaterial((JPatchMaterial) it.next());
			addEdit(new AddPatchEdit(patch));
		}
	}
	
	/**
	* creates a new cloned ControlPoint and adds it to the maps
	**/
	protected void addToCloneMap(ControlPoint cp) {
		ControlPoint clone = new ControlPoint(cp);
		mapClones.put(cp,clone);
		mapOriginals.put(clone,cp);
	}

	/**
	* returns the original controlPoint for a clone
	**/
	protected ControlPoint getOriginal(ControlPoint clone) {
		return (ControlPoint) mapOriginals.get(clone);
	}
	
	/**
	* returns the clone for a original controlPoint
	**/
	protected ControlPoint getClone(ControlPoint original) {
		return (ControlPoint) mapClones.get(original);
	}
	
	/**
	* checks if a direct neighbot is part of the selection
	**/
	protected boolean checkForNeighbor(ControlPoint cp) {
		return ((cp.getNext() != null && JPatchUtils.arrayContains(acp,cp.getNext().getHead())) || (cp.getPrev() != null && JPatchUtils.arrayContains(acp,cp.getPrev().getHead())));
	}

	/**
	* checks if start cp of a hook curve should be added
	**/
	protected boolean checkStartHook(ControlPoint cp) {

		/* check if end of hook-curve is part of the selection */
		if (JPatchUtils.arrayContains(acp, cp.getEnd().getHead())) {

			/* we need at least one more selected point on the hook curve */
			/* so let's search for one... */
			for (ControlPoint cpHookCurve = cp; cpHookCurve.getNext() != null; cpHookCurve = cpHookCurve.getNext()) {

				/* check if the point and its direct neighbor are part of the selection too */
				if (JPatchUtils.arrayContains(acp,cpHookCurve) && cpHookCurve.getPrevAttached() != null && checkForNeighbor(cpHookCurve.getPrevAttached())) {
					return true;
				}
			}
		}
		return false;
	}
}

