/*
 * $Id$
 *
 * Copyright (c) 2004 Sascha Ledinsky
 *
 * This file is part of JPatch.
 * 
 * JPatch is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * JPatch is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Foobar; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package net.sf.jpatch.entity;

import java.util.*;
import javax.vecmath.*;
import net.sf.jpatch.renderer.*;
import net.sf.jpatch.boundary.*;
import net.sf.jpatch.auxilary.*;

/**
 * This class contains a material-name, materialProperties and renderStrings.
 * It provides methods which returns string representations of the material
 * which can be used for rendering (POV-Ray, RenderMAN) or XML export.
 * <p>
 * The following variables may be used in renderStrings and will be substituted
 * by the materialProperty values:
 * <table>
 * <tr><th align="left">$name</th><td>The name of the material</td></tr>
 * <tr><th align="left">$r</th><td>The red component of the materials color 0..1</td></tr>
 * <tr><th align="left">$g</th><td>The green component of the materials color 0..1</td></tr>
 * <tr><th align="left">$b</th><td>The blue component of the materials color 0..1</td></tr>
 * <tr><th align="left">$filter</th><td>The amount of light which is being filtered through the material 0..1</td></tr>
 * <tr><th align="left">$transmit</th><td>The amount of light which will pass unfiltered through the material 0..1</td></tr>
 * <tr><th align="left">$opacity</th><td>The opacity values (RGB) of the material, used by RenderMAN shaders</td></tr>
 * <tr><th align="left">$ambient</th><td>The amount of ambient light 0..1</td></tr>
 * <tr><th align="left">$diffuse</th><td>The amount of diffuse light 0..1</td></tr>
 * <tr><th align="left">$brilliance</th><td>The brilliance value 0..&infin;</td></tr>
 * <tr><th align="left">$specular</th><td>The amount of specular light 0..&infin;</td></tr>
 * <tr><th align="left">$roughness</th><td>The roughness value 0..1</td></tr>
 * <tr><th align="left">$metallic</th><td>The metallic value 0..1</td></tr>
 * <tr><th align="left">$reflection</th><td>The amount of reflection 0..1</td></tr>
 * <tr><th align="left">$reflection_min</th><td>The minimum reflection (used for variable reflection) 0..reflection_max</td></tr>
 * <tr><th align="left">$reflection_max</th><td>The maximum reflection (used for variable reflection) reflection_min..1</td></tr>
 * <tr><th align="left">$reflection_falloff</th><td>The exponent of variable reflection 0..&infin;</td></tr>
 * <tr><th align="left">$refraction</th><td>The index of refraction 1..&infin;</td></tr>
 * <tr><th align="left">$conserve_energy</th><td>The converse energy flag ture|false</td></tr>
 * </table>
 *
 * @author     Sascha Ledinsky
 * @version    $Revision$
 * @see		net.sf.jpatch.entity.MaterialProperties
 */

public class JPatchMaterial extends JPatchTreeLeaf {
	/** contains the default renderStrings */
	private static final Map mapDefault = new HashMap();

	/** materialProperties */
	private MaterialProperties materialProperties = new MaterialProperties();
	/** a map of renderStrings */
	private HashMap mapRenderStrings = new HashMap();

	private int iXmlNumber;
	/**
	 * set default renderStrings
	 */
	static {
		mapDefault.put("povray","pigment {\n\tcolor rgbft <$r,$g,$b,$filter,$transmit>\n}\nfinish {\n\tambient $ambient\n\tdiffuse $diffuse brilliance $brilliance\n\tspecular $specular roughness $roughness metallic $metallic\n\t#if ($conserve_energy) conserve_energy #end\n\treflection {\n\t\t$reflection_min,$reflection_max\n\t\tfalloff $reflection_falloff\n\t\tmetallic $metallic\n\t}\n}\ninterior {\n\tior $refraction\n}\n");
		//mapDefault.put("renderman","Color [$r $g $b]\nOpacity [$opacity]\nSurface \"jpatch_default\" \"Ka\" [$ambient] \"Kd\" [$diffuse] \"brilliance\" [$brilliance] \"Ks\" [$specular] \"roughness\" [$roughness] \"metallic\" [$metallic] \"conserve_energy\" [$conserve_energy] \"Kr\" [$reflection_min] \"Kr_var\" [$reflection_variance] \"Kr_falloff\" [$reflection_falloff] \"Kt\" [$refraction]\n");
		mapDefault.put("renderman","Color [$r $g $b]\nOpacity [$opacity]\nSurface \"plastic\" \"Ka\" [$ambient] \"Kd\" [$diffuse] \"Ks\" [$specular] \"roughness\" [$roughness]\n");
	}

	/**
	 * Default Constructor
	 */
	public JPatchMaterial() {
		strName = "New Material";
		iNodeType = MATERIAL;
		setRenderString("povray","3.5",(String)mapDefault.get("povray"));
		setRenderString("renderman","3.1",(String)mapDefault.get("renderman"));
	}

	/**
	 * Constructor
	 *
	 * @param color	The color of the material
	 */
	public JPatchMaterial(Color3f color) {
		this();
		materialProperties.red = color.x;
		materialProperties.green = color.y;
		materialProperties.blue = color.z;
	}

	/**
	 * Sets the renderstring for a renderer
	 *
	 * @param format	The format. Possible values are "povray" or "renderman"
	 * @param version	The version. Not used at the moment
	 * @param renderString	The renderString
	 */
	public void setRenderString(String format, String version, String renderString) {
		String key = format;
		if (mapRenderStrings.containsKey(key)) {
			mapRenderStrings.remove(key);
		}
		mapRenderStrings.put(key,renderString);
	}

	/**
	 * gets the renderString for a renderer format
	 *
	 * @param format	The format. Possible values are "povray" or "renderman"
	 * @param version	The version. Not used at the moment
	 */
	public String getRenderString(String format, String version) {
		String key = format;
		if (mapRenderStrings.containsKey(key)) {
			return (String)mapRenderStrings.get(key);
		}
		return null;
	}

	/**
	 * sets the color of the material
	 *
	 * @param color The color to set
	 */
	public void setColor(Color3f color) {
		materialProperties.red = color.x;
		materialProperties.green = color.y;
		materialProperties.blue = color.z;
	}

	
	public void setXmlNumber(int n) {
		iXmlNumber = n;
	}
	
	public int getXmlNumber() {
		return iXmlNumber;
	}
	
	/**
	 * used by xml method
	 */
	public StringBuffer renderStrings(int tabs) {
		StringBuffer sbIndent = XMLutils.indent(tabs);
		StringBuffer sbLineBreak = XMLutils.lineBreak();
		StringBuffer sb = new StringBuffer();
		Iterator iterator = mapRenderStrings.keySet().iterator();
		while (iterator.hasNext()) {
			String key = (String)iterator.next();
			String renderString = (String)mapRenderStrings.get(key);
			if (!renderString.equals((String)mapDefault.get(key))) {
				String r;
				r = renderString.replaceAll("&","&amp;");
				r = r.replaceAll("\"","&quot;");
				r = r.replaceAll(">","&gt;");
				r = r.replaceAll("<","&lt;");
				sb.append(sbIndent).append("<renderer format=").append(XMLutils.quote(key)).append(">");
				sb.append(r);
				sb.append(sbIndent).append("</renderer>").append(sbLineBreak);
			}
		}
		return sb;
	}

	/**
	 * Get XML representation of material
	 *
	 * @param tabs The indent level
	 */
	public StringBuffer xml(int tabs) {
		StringBuffer sbIndent = XMLutils.indent(tabs);
		StringBuffer sbLineBreak = XMLutils.lineBreak();
		StringBuffer sb = new StringBuffer();
		sb.append(sbIndent).append("<material>").append(sbLineBreak);
		sb.append(sbIndent).append("\t<name>").append(strName).append("</name>").append(sbLineBreak);
		sb.append(materialProperties.xml(tabs + 1));
		sb.append(renderStrings(tabs + 1));
		sb.append(sbIndent).append("</material>").append(sbLineBreak);
		return sb;
	}

	/**
	 * Get MaterialProperties of this Material
	 *
	 * @return The MatreialProperties of this Material
	 */
	public MaterialProperties getMaterialProperties() {
		return materialProperties;
	}

	/**
	 * Replaces the strings like $name, $r, $g, $b,...
	 */
	private String replaceString(String str) {
		String opacity = materialProperties.opacityRed() + " " + materialProperties.opacityGreen() + " " + materialProperties.opacityBlue();
		String s = str;
		s = s.replaceAll("\\$name",strName);
		s = s.replaceAll("\\$r",String.valueOf(materialProperties.red));
		s = s.replaceAll("\\$g",String.valueOf(materialProperties.green));
		s = s.replaceAll("\\$b",String.valueOf(materialProperties.blue));
		s = s.replaceAll("\\$filter",String.valueOf(materialProperties.filter));
		s = s.replaceAll("\\$transmit",String.valueOf(materialProperties.transmit));
		s = s.replaceAll("\\$opacity",String.valueOf(opacity));
		s = s.replaceAll("\\$ambient",String.valueOf(materialProperties.ambient));
		s = s.replaceAll("\\$diffuse",String.valueOf(materialProperties.diffuse));
		s = s.replaceAll("\\$brilliance",String.valueOf(materialProperties.diffuse));
		s = s.replaceAll("\\$specular",String.valueOf(materialProperties.specular));
		s = s.replaceAll("\\$roughness",String.valueOf(materialProperties.roughness));
		s = s.replaceAll("\\$metallic",String.valueOf(materialProperties.metallic));
		s = s.replaceAll("\\$reflection_min",String.valueOf(materialProperties.reflectionMin));
		s = s.replaceAll("\\$reflection_max",String.valueOf(materialProperties.reflectionMax));
		s = s.replaceAll("\\$reflection_falloff",String.valueOf(materialProperties.reflectionFalloff));
		s = s.replaceAll("\\$reflection",String.valueOf(materialProperties.reflection()));
		s = s.replaceAll("\\$refraction",String.valueOf(materialProperties.refraction));
		s = s.replaceAll("\\$conserve_energy",String.valueOf(materialProperties.conserveEnergy));
		return s;
	}
}
