/*
 * $Id$
 *
 * Copyright (c) 2004 Sascha Ledinsky
 *
 * This file is part of JPatch.
 * 
 * JPatch is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * JPatch is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Foobar; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package net.sf.jpatch.entity;

import javax.vecmath.*;
import net.sf.jpatch.boundary.*;
import net.sf.jpatch.boundary.selection.*;
import net.sf.jpatch.auxilary.*;
import java.io.*;

/**
 * A Patch. Currently 3-, 4- and 5-point patches are supported
 *
 * @author     Sascha Ledinsky
 * @version    $Revision$
 * @see		net.sf.jpatch.entity.Curve
 * @see		net.sf.jpatch.entity.ControlPoint
 */
public final class Patch {
	private Patch ptNext;
	private Patch ptPrev;
	private ControlPoint[] acpPoint;
	//private boolean bReTriangulize = true;
	//private TriangulizedPatch triangulizedPatch;
	private JPatchMaterial material = MainFrame.getInstance().getModel().getMaterial(0);//new JPatchMaterial();
	private boolean bDynamic = false;
	private boolean bValid = true;
	private Model model;
	
	public Patch(ControlPoint[] controlPoints) {
		acpPoint = controlPoints;
		//setGroup();
		//reset();
		//for (int i = 0; i < acpPoint.length; i++) {
		//	acpPoint[i].addPatch(this);
		//}
	}
	
	public final boolean contains(ControlPoint cp) {
		for (int n = 0; n < acpPoint.length; n++) {
			if (cp == acpPoint[n]) return true;
		}
		return false;
	}
	
	public ControlPoint[] getControlPoints() {
		return acpPoint;
	}
	
	public String toString() {
		String s = "" + hashCode() + " ";
		for (int n = 0; n < acpPoint.length; n++) {
			s += acpPoint[n].toString() + " ";
		}
		return s;
	}
	
	public final boolean isEqual(ControlPoint[] acp) {
		/*
		 * check for equal size
		 */
		if (acpPoint.length == acp.length) {
			boolean result = false;
			int l = acpPoint.length;
			for (int n = 0; n < l; n++) {
				if (acpPoint[0] == acp[n]) {
					if (acpPoint[1] == acp[(n + 1) % l]) {
						for (int m = 2; m < l; m++) {
							if (acpPoint[m] != acp[(n + m) % l]) {
								return false;
							}
						}
						return true;
					} else if (acpPoint[1] == acp[(n + l - 1) % l]) {
						for (int m = 2; m < l; m++) {
							if (acpPoint[m] != acp[(n + l - m) % l]) {
								return false;
							}
						}
						return true;
					} else return false;
				}
			}
		}
		return false;
	}

	public final void check(PointSelection ps) {
		bDynamic = false;
		loop:
		for (int i = 0; i < acpPoint.length; i++) {
			if (ps.contains(acpPoint[i].getHead())) bDynamic = true;
			if (acpPoint[i].getPrev() != null && ps.contains(acpPoint[i].getPrev().getHead())) bDynamic = true;
			if (acpPoint[i].getNext() != null && ps.contains(acpPoint[i].getNext().getHead())) bDynamic = true;
			if (bDynamic) break loop;
		}
	}
	
	public final boolean isSelected(PointSelection ps) {
		for (int i = 0; i < acpPoint.length; i++) {
			if (!ps.contains(acpPoint[i].getHead())) {
				if (!acpPoint[i].isTargetHook() && !acpPoint[i].isChildHook()) return false;
			}
		}
		return true;
	}
	
	public final boolean isValid() {
		return bValid;
	}
	
	public final void setValid(boolean valid) {
		bValid = valid;
	}
	
	public final boolean isDynamic() {
		return bDynamic;
	}
	/*
	public final void reset() {
		triangulizedPatch = MainFrame.getInstance().getDisplayFactory().createTriangulizedPatch(this);
	}
	
	public final TriangulizedPatch getTriangulizedPatch() {
		return triangulizedPatch;
	}
	*/
	public final void setNext(Patch next) {
		ptNext = next;
	}
	
	public final void setPrev(Patch prev) {
		ptPrev = prev;
	}
	
	public final Patch getNext() {
		return ptNext;
	}
	
	public final Patch getPrev() {
		return ptPrev;
	}
	
	public final int getType() {
		int l = acpPoint.length;
		if (l == 6 || l == 8 || l == 10) {
			return l / 2;
		} else {
			return -1;
		}
	}
	
	public final int getSize() {
		switch(getType()) {
			case 3:
				return 1;
			case 4:
				return 1;
			case 5:
				return 5;
		}
		return -1;
	}
	
	public final void remove() {
		if (ptPrev != null) {
			ptPrev.ptNext = ptNext;
		}
		if (ptNext != null) {
			ptNext.ptPrev = ptPrev;
		}
		if (model != null && model.getLastPatch() == this) {
			model.setLastPatch(ptPrev);
		}
		if (model != null && model.getFirstPatch() == this) {
			model.setFirstPatch(ptNext);
		}
		model = null;
	}

	public Model getModel() {
		return model;
	}
	
	public void setModel(Model model) {
		this.model = model;
	}
	
	/**
	 * returns StringBuffer containing an XML representation of the curve, used to save
	 * models in XML format
	 *
	 * @return A StringBuffer containing an XML representation of this curve
	 */
	public StringBuffer xml(int tabs) {
		StringBuffer sbIndent = XMLutils.indent(tabs);
		StringBuffer sbIndent1 = XMLutils.indent(tabs + 1);
		StringBuffer sbLineBreak = XMLutils.lineBreak();
		StringBuffer sb = new StringBuffer();
		sb.append(sbIndent).append("<patch material=\"" + material.getXmlNumber() + "\">").append(sbLineBreak);
		sb.append(sbIndent1).append("<points>");
		//int size = getType();
		//int p = 0;
		for (int i = 0; i < acpPoint.length; i++) {
			sb.append(acpPoint[i].getXmlNumber());
			if (i < acpPoint.length - 1) {
				sb.append(",");
			}
		}
		sb.append("</points>").append(sbLineBreak);
		sb.append(sbIndent).append("</patch>").append(sbLineBreak);
		return sb;
	}
	
	private Point3f parallelogram(Point3f a, Point3f b, Point3f c) {
		Point3f result = new Point3f(b);
		result.add(c);
		result.sub(a);
		return result;
	}
	
	public Point3f[] coonsPatch() {
		Point3f[]ap3CoonsPatch = new Point3f[12];
		if (acpPoint.length == 6) {
			ap3CoonsPatch = new Point3f[9];
		}
		if (acpPoint.length == 8) {
			ap3CoonsPatch = new Point3f[12];
		}
		if (acpPoint.length == 10) {
			ap3CoonsPatch = new Point3f[15];
		}
		int coon = 0;
		int p = 0;
		while (p < acpPoint.length) {
			ap3CoonsPatch[coon++] = new Point3f(acpPoint[p].getPosition());
			if (acpPoint[p].getNext() == acpPoint[p + 1]) {
				ap3CoonsPatch[coon++] = new Point3f(acpPoint[p].getOutTangent());
				ap3CoonsPatch[coon++] = new Point3f(acpPoint[++p].getInTangent());
			} else {
				ap3CoonsPatch[coon++] = new Point3f(acpPoint[p].getInTangent());
				ap3CoonsPatch[coon++] = new Point3f(acpPoint[++p].getOutTangent());
			}
			p++;
		}
		return ap3CoonsPatch;
	}
	
	public void setMaterial(JPatchMaterial material) {
		this.material = material;
	}
	
	public JPatchMaterial getMaterial() {
		return material;
	}

	public Point3f[][] bicubicPatches() {
		Point3f[][] aap3Patches = null;
		Point3f[] ap3CoonsPatch = coonsPatch();
		Vector3f v3A = new Vector3f();
		Vector3f v3B = new Vector3f();
		Vector3f v3C = new Vector3f();
		Point3f p3cp5 = new Point3f();
		Point3f p3cp6 = new Point3f();
		Point3f p3cp9 = new Point3f();
		Point3f p3cp10 = new Point3f();
		switch(acpPoint.length) {
			case 10:
				
				aap3Patches = new Point3f[5][16];
				Point3f[] ap3 = new Point3f[41];
				Point3f p3_helper1 = new Point3f();
				Point3f p3_helper2 = new Point3f();
				Point3f[] ap3Helper = new Point3f[7];
				
				Bezier.deCasteljau(ap3CoonsPatch[0],ap3CoonsPatch[1],ap3CoonsPatch[2],ap3CoonsPatch[3],0.5f,ap3,0);
				Bezier.deCasteljau(ap3CoonsPatch[3],ap3CoonsPatch[4],ap3CoonsPatch[5],ap3CoonsPatch[6],0.5f,ap3,6);
				Bezier.deCasteljau(ap3CoonsPatch[6],ap3CoonsPatch[7],ap3CoonsPatch[8],ap3CoonsPatch[9],0.5f,ap3,12);
				Bezier.deCasteljau(ap3CoonsPatch[9],ap3CoonsPatch[10],ap3CoonsPatch[11],ap3CoonsPatch[12],0.5f,ap3,18);
				Bezier.deCasteljau(ap3CoonsPatch[12],ap3CoonsPatch[13],ap3CoonsPatch[14],ap3CoonsPatch[0],0.5f,ap3,24);
				
				p3_helper1.interpolate(ap3CoonsPatch[2],ap3CoonsPatch[13],0.5f);
				p3_helper2.interpolate(ap3CoonsPatch[4],ap3CoonsPatch[11],0.5f);
				ap3Helper = Bezier.deCasteljau(ap3[0],p3_helper1,p3_helper2,ap3[15],3f/5f);
				ap3[30] = ap3Helper[3];
				ap3[31] = ap3Helper[4];
				//ap3[32] = ap3Helper[5];
				
				p3_helper1.interpolate(ap3CoonsPatch[1],ap3CoonsPatch[5],0.5f);
				p3_helper2.interpolate(ap3CoonsPatch[14],ap3CoonsPatch[7],0.5f);
				ap3Helper = Bezier.deCasteljau(ap3[6],p3_helper1,p3_helper2,ap3[21],3f/5f);
				ap3[30].add(ap3Helper[3]);
				ap3[33] = ap3Helper[4];
				//ap3[34] = ap3Helper[5];
				
				p3_helper1.interpolate(ap3CoonsPatch[4],ap3CoonsPatch[8],0.5f);
				p3_helper2.interpolate(ap3CoonsPatch[2],ap3CoonsPatch[10],0.5f);
				ap3Helper = Bezier.deCasteljau(ap3[12],p3_helper1,p3_helper2,ap3[27],3f/5f);
				ap3[30].add(ap3Helper[3]);
				ap3[35] = ap3Helper[4];
				//ap3[36] = ap3Helper[5];
				
				p3_helper1.interpolate(ap3CoonsPatch[7],ap3CoonsPatch[11],0.5f);
				p3_helper2.interpolate(ap3CoonsPatch[5],ap3CoonsPatch[13],0.5f);
				ap3Helper = Bezier.deCasteljau(ap3[18],p3_helper1,p3_helper2,ap3[3],3f/5f);
				ap3[30].add(ap3Helper[3]);
				ap3[37] = ap3Helper[4];
				//ap3[38] = ap3Helper[5];
				
				p3_helper1.interpolate(ap3CoonsPatch[10],ap3CoonsPatch[14],0.5f);
				p3_helper2.interpolate(ap3CoonsPatch[1],ap3CoonsPatch[8],0.5f);
				ap3Helper = Bezier.deCasteljau(ap3[24],p3_helper1,p3_helper2,ap3[9],3f/5f);
				ap3[30].add(ap3Helper[3]);
				ap3[39] = ap3Helper[4];
				//ap3[40] = ap3Helper[5];
				
				ap3[30].scale(1f/5f);
				
				Vector3f v3_n0 = new Vector3f();
				Vector3f v3_n1 = new Vector3f();
				Vector3f v3_n2 = new Vector3f();
				Vector3f v3_n3 = new Vector3f();
				Vector3f v3_n4 = new Vector3f();
				
				v3_n0.cross(Functions.vector(ap3[30],ap3[35]),Functions.vector(ap3[30],ap3[37]));
				v3_n1.cross(Functions.vector(ap3[30],ap3[37]),Functions.vector(ap3[30],ap3[39]));
				v3_n2.cross(Functions.vector(ap3[30],ap3[39]),Functions.vector(ap3[30],ap3[31]));
				v3_n3.cross(Functions.vector(ap3[30],ap3[31]),Functions.vector(ap3[30],ap3[33]));
				v3_n4.cross(Functions.vector(ap3[30],ap3[33]),Functions.vector(ap3[30],ap3[35]));
				
				Vector3f v3Normal = new Vector3f(v3_n0);
				v3Normal.add(v3_n1);
				v3Normal.add(v3_n2);
				v3Normal.add(v3_n3);
				v3Normal.add(v3_n4);
				v3Normal.normalize();
				
				Plane plane = new Plane(ap3[30],v3Normal);
				
				ap3[31].set(plane.projectedPoint(ap3[31]));
				ap3[33].set(plane.projectedPoint(ap3[33]));
				ap3[35].set(plane.projectedPoint(ap3[35]));
				ap3[37].set(plane.projectedPoint(ap3[37]));
				ap3[39].set(plane.projectedPoint(ap3[39]));
				
				Vector3f v3_3 = new Vector3f();
				v3_3.interpolate(Functions.vector(ap3[0],ap3[29]),Functions.vector(ap3[6],ap3[7]),0.5f);
				Vector3f v3_9 = new Vector3f();
				v3_9.interpolate(Functions.vector(ap3[6],ap3[5]),Functions.vector(ap3[12],ap3[13]),0.5f);
				Vector3f v3_15 = new Vector3f();
				v3_15.interpolate(Functions.vector(ap3[12],ap3[11]),Functions.vector(ap3[18],ap3[19]),0.5f);
				Vector3f v3_21 = new Vector3f();
				v3_21.interpolate(Functions.vector(ap3[18],ap3[17]),Functions.vector(ap3[24],ap3[25]),0.5f);
				Vector3f v3_27 = new Vector3f();
				v3_27.interpolate(Functions.vector(ap3[24],ap3[23]),Functions.vector(ap3[0],ap3[1]),0.5f);
				
				ap3[38] = new Point3f(ap3[3]);
				ap3[38].add(v3_3);
				ap3[40] = new Point3f(ap3[9]);
				ap3[40].add(v3_9);
				ap3[32] = new Point3f(ap3[15]);
				ap3[32].add(v3_15);
				ap3[34] = new Point3f(ap3[21]);
				ap3[34].add(v3_21);
				ap3[36] = new Point3f(ap3[27]);
				ap3[36].add(v3_27);
				
				
				aap3Patches[0][0] = new Point3f(ap3[0]);
				aap3Patches[0][1] = new Point3f(ap3[1]);
				aap3Patches[0][2] = new Point3f(ap3[2]);
				aap3Patches[0][3] = new Point3f(ap3[3]);
				aap3Patches[0][7] = new Point3f(ap3[38]);
				aap3Patches[0][11] = new Point3f(ap3[37]);
				aap3Patches[0][15] = new Point3f(ap3[30]);
				aap3Patches[0][14] = new Point3f(ap3[35]);
				aap3Patches[0][13] = new Point3f(ap3[36]);
				aap3Patches[0][12] = new Point3f(ap3[27]);
				aap3Patches[0][8] = new Point3f(ap3[28]);
				aap3Patches[0][4] = new Point3f(ap3[29]);
				
				aap3Patches[1][0] = new Point3f(ap3[6]);
				aap3Patches[1][1] = new Point3f(ap3[7]);
				aap3Patches[1][2] = new Point3f(ap3[8]);
				aap3Patches[1][3] = new Point3f(ap3[9]);
				aap3Patches[1][7] = new Point3f(ap3[40]);
				aap3Patches[1][11] = new Point3f(ap3[39]);
				aap3Patches[1][15] = new Point3f(ap3[30]);
				aap3Patches[1][14] = new Point3f(ap3[37]);
				aap3Patches[1][13] = new Point3f(ap3[38]);
				aap3Patches[1][12] = new Point3f(ap3[3]);
				aap3Patches[1][8] = new Point3f(ap3[4]);
				aap3Patches[1][4] = new Point3f(ap3[5]);
				
				aap3Patches[2][0] = new Point3f(ap3[12]);
				aap3Patches[2][1] = new Point3f(ap3[13]);
				aap3Patches[2][2] = new Point3f(ap3[14]);
				aap3Patches[2][3] = new Point3f(ap3[15]);
				aap3Patches[2][7] = new Point3f(ap3[32]);
				aap3Patches[2][11] = new Point3f(ap3[31]);
				aap3Patches[2][15] = new Point3f(ap3[30]);
				aap3Patches[2][14] = new Point3f(ap3[39]);
				aap3Patches[2][13] = new Point3f(ap3[40]);
				aap3Patches[2][12] = new Point3f(ap3[9]);
				aap3Patches[2][8] = new Point3f(ap3[10]);
				aap3Patches[2][4] = new Point3f(ap3[11]);
				
				aap3Patches[3][0] = new Point3f(ap3[18]);
				aap3Patches[3][1] = new Point3f(ap3[19]);
				aap3Patches[3][2] = new Point3f(ap3[20]);
				aap3Patches[3][3] = new Point3f(ap3[21]);
				aap3Patches[3][7] = new Point3f(ap3[34]);
				aap3Patches[3][11] = new Point3f(ap3[33]);
				aap3Patches[3][15] = new Point3f(ap3[30]);
				aap3Patches[3][14] = new Point3f(ap3[31]);
				aap3Patches[3][13] = new Point3f(ap3[32]);
				aap3Patches[3][12] = new Point3f(ap3[15]);
				aap3Patches[3][8] = new Point3f(ap3[16]);
				aap3Patches[3][4] = new Point3f(ap3[17]);
				
				aap3Patches[4][0] = new Point3f(ap3[24]);
				aap3Patches[4][1] = new Point3f(ap3[25]);
				aap3Patches[4][2] = new Point3f(ap3[26]);
				aap3Patches[4][3] = new Point3f(ap3[27]);
				aap3Patches[4][7] = new Point3f(ap3[36]);
				aap3Patches[4][11] = new Point3f(ap3[35]);
				aap3Patches[4][15] = new Point3f(ap3[30]);
				aap3Patches[4][14] = new Point3f(ap3[33]);
				aap3Patches[4][13] = new Point3f(ap3[34]);
				aap3Patches[4][12] = new Point3f(ap3[21]);
				aap3Patches[4][8] = new Point3f(ap3[22]);
				aap3Patches[4][4] = new Point3f(ap3[23]);
				
				BezierPatch.computeInnerControlPoints(aap3Patches[0]);
				BezierPatch.computeInnerControlPoints(aap3Patches[1]);
				BezierPatch.computeInnerControlPoints(aap3Patches[2]);
				BezierPatch.computeInnerControlPoints(aap3Patches[3]);
				BezierPatch.computeInnerControlPoints(aap3Patches[4]);
				
				/*
				aap3Patches[0][0] = ap3CoonsPatch[0];
				aap3Patches[0][1] = ap3CoonsPatch[1];
				aap3Patches[0][2] = ap3CoonsPatch[2];
				aap3Patches[0][3] = ap3CoonsPatch[3];
				aap3Patches[0][7] = ap3CoonsPatch[4];
				aap3Patches[0][11] = ap3CoonsPatch[5];
				aap3Patches[0][15] = ap3CoonsPatch[6];
				aap3Patches[0][14] = ap3CoonsPatch[7];
				aap3Patches[0][13] = ap3CoonsPatch[8];
				aap3Patches[0][12] = ap3CoonsPatch[9];
				aap3Patches[0][8] = ap3CoonsPatch[10];
				aap3Patches[0][4] = ap3CoonsPatch[11];
				/*
				aap3Patches[0][5] = parallelogram(aap3Patches[0][0],aap3Patches[0][1],aap3Patches[0][4]);
				aap3Patches[0][6] = parallelogram(aap3Patches[0][3],aap3Patches[0][2],aap3Patches[0][7]);
				aap3Patches[0][9] = parallelogram(aap3Patches[0][12],aap3Patches[0][8],aap3Patches[0][13]);
				aap3Patches[0][10] = parallelogram(aap3Patches[0][15],aap3Patches[0][11],aap3Patches[0][14]);
				
				BezierPatch.computeInnerControlPoints(aap3Patches[0]);
				*/
				/*
				Point3f[] ap3Helper;
				Point3f p3_1 = new Point3f();
				Point3f p3_2 = new Point3f();
				p3_1.interpolate(ap3CoonsPatch[2],ap3CoonsPatch[13],0.5f);
				p3_2.interpolate(ap3CoonsPatch[4],ap3CoonsPatch[11],0.5f);
				ap3Helper = Bezier.deCasteljau(ap3CoonsPatch[9],ap3CoonsPatch[8],ap3CoonsPatch[7],ap3CoonsPatch[6],0.5f);
				
				aap3Patches[0][0] = ap3CoonsPatch[12];
				aap3Patches[0][1] = ap3CoonsPatch[13];
				aap3Patches[0][2] = ap3CoonsPatch[14];
				aap3Patches[0][3] = ap3CoonsPatch[0];
				aap3Patches[0][7] = new Point3f(p3_1);
				aap3Patches[0][11] = new Point3f(p3_2);
				aap3Patches[0][15] = new Point3f(ap3Helper[3]);
				aap3Patches[0][14] = ap3Helper[2];
				aap3Patches[0][13] = ap3Helper[1];
				aap3Patches[0][12] = ap3Helper[0];
				aap3Patches[0][8] = ap3CoonsPatch[10];
				aap3Patches[0][4] = ap3CoonsPatch[11];
				BezierPatch.computeInnerControlPoints(aap3Patches[0]);
				
				aap3Patches[1][0] = new Point3f(ap3CoonsPatch[0]);
				aap3Patches[1][1] = ap3CoonsPatch[1];
				aap3Patches[1][2] = ap3CoonsPatch[2];
				aap3Patches[1][3] = ap3CoonsPatch[3];
				aap3Patches[1][7] = ap3CoonsPatch[4];
				aap3Patches[1][11] = ap3CoonsPatch[5];
				aap3Patches[1][15] = ap3Helper[6];
				aap3Patches[1][14] = ap3Helper[5];
				aap3Patches[1][13] = ap3Helper[4];
				aap3Patches[1][12] = ap3Helper[3];
				aap3Patches[1][8] = p3_2;
				aap3Patches[1][4] = p3_1;
				BezierPatch.computeInnerControlPoints(aap3Patches[1]);
				*/
			break;
			case 8:
				aap3Patches = new Point3f[1][16];
				aap3Patches[0][0] = ap3CoonsPatch[0];
				aap3Patches[0][1] = ap3CoonsPatch[1];
				aap3Patches[0][2] = ap3CoonsPatch[2];
				aap3Patches[0][3] = ap3CoonsPatch[3];
				aap3Patches[0][7] = ap3CoonsPatch[4];
				aap3Patches[0][11] = ap3CoonsPatch[5];
				aap3Patches[0][15] = ap3CoonsPatch[6];
				aap3Patches[0][14] = ap3CoonsPatch[7];
				aap3Patches[0][13] = ap3CoonsPatch[8];
				aap3Patches[0][12] = ap3CoonsPatch[9];
				aap3Patches[0][8] = ap3CoonsPatch[10];
				aap3Patches[0][4] = ap3CoonsPatch[11];
				/*
				aap3Patches[0][5] = parallelogram(aap3Patches[0][0],aap3Patches[0][1],aap3Patches[0][4]);
				aap3Patches[0][6] = parallelogram(aap3Patches[0][3],aap3Patches[0][2],aap3Patches[0][7]);
				aap3Patches[0][9] = parallelogram(aap3Patches[0][12],aap3Patches[0][8],aap3Patches[0][13]);
				aap3Patches[0][10] = parallelogram(aap3Patches[0][15],aap3Patches[0][11],aap3Patches[0][14]);
				*/
				BezierPatch.computeInnerControlPoints(aap3Patches[0]);
				
			break;
			case 6:
				
				aap3Patches = new Point3f[1][16];
				aap3Patches[0][0] = ap3CoonsPatch[0];
				aap3Patches[0][1] = ap3CoonsPatch[1];
				aap3Patches[0][2] = ap3CoonsPatch[2];
				aap3Patches[0][3] = ap3CoonsPatch[3];
				aap3Patches[0][7] = ap3CoonsPatch[4];
				aap3Patches[0][11] = ap3CoonsPatch[5];
				/*
				aap3Patches[0][15] = ap3CoonsPatch[6];
				aap3Patches[0][14] = new Point3f(ap3CoonsPatch[6]);
				aap3Patches[0][13] = new Point3f(ap3CoonsPatch[6]);
				aap3Patches[0][12] = new Point3f(ap3CoonsPatch[6]);
				*/
				aap3Patches[0][15] = Bezier.evaluate(ap3CoonsPatch[3],ap3CoonsPatch[4],ap3CoonsPatch[5],ap3CoonsPatch[6],0.9999f);
				aap3Patches[0][12] = Bezier.evaluate(ap3CoonsPatch[0],ap3CoonsPatch[8],ap3CoonsPatch[7],ap3CoonsPatch[6],0.9999f);
				aap3Patches[0][13] = new Point3f();
				aap3Patches[0][14] = new Point3f();
				aap3Patches[0][13].interpolate(aap3Patches[0][12],aap3Patches[0][15],1f/3f);
				aap3Patches[0][14].interpolate(aap3Patches[0][12],aap3Patches[0][15],2f/3f);
				
				aap3Patches[0][8] = ap3CoonsPatch[7];
				aap3Patches[0][4] = ap3CoonsPatch[8];
				BezierPatch.computeInnerControlPoints(aap3Patches[0]);
				/*
				aap3Patches[0][5] = parallelogram(aap3Patches[0][0],aap3Patches[0][1],aap3Patches[0][4]);
				aap3Patches[0][6] = parallelogram(aap3Patches[0][3],aap3Patches[0][2],aap3Patches[0][7]);
				aap3Patches[0][9] = parallelogram(aap3Patches[0][12],aap3Patches[0][8],aap3Patches[0][13]);
				aap3Patches[0][10] = parallelogram(aap3Patches[0][15],aap3Patches[0][11],aap3Patches[0][14]);
				
				Point3f[] ap3Helper;
				Vector3f v3_a;
				Vector3f v3_b;
				Vector3f v3_3 = new Vector3f();
				Vector3f v3_9 = new Vector3f();
				Vector3f v3_15 = new Vector3f();
				
				ap3Helper = Bezier.deCasteljau(ap3CoonsPatch[0],ap3CoonsPatch[1],ap3CoonsPatch[2],ap3CoonsPatch[3],0.5f);
				Point3f p3_0 = ap3Helper[0];
				Point3f p3_1 = ap3Helper[1];
				Point3f p3_2 = ap3Helper[2];
				Point3f p3_3 = ap3Helper[3];
				Point3f p3_4 = ap3Helper[4];
				Point3f p3_5 = ap3Helper[5];
				Point3f p3_6 = ap3Helper[6];
				ap3Helper = Bezier.deCasteljau(ap3CoonsPatch[3],ap3CoonsPatch[4],ap3CoonsPatch[5],ap3CoonsPatch[6],0.5f);
				Point3f p3_7 = ap3Helper[1];
				Point3f p3_8 = ap3Helper[2];
				Point3f p3_9 = ap3Helper[3];
				Point3f p3_10 = ap3Helper[4];
				Point3f p3_11 = ap3Helper[5];
				Point3f p3_12 = ap3Helper[6];
				ap3Helper = Bezier.deCasteljau(ap3CoonsPatch[6],ap3CoonsPatch[7],ap3CoonsPatch[8],ap3CoonsPatch[0],0.5f);
				Point3f p3_13 = ap3Helper[1];
				Point3f p3_14 = ap3Helper[2];
				Point3f p3_15 = ap3Helper[3];
				Point3f p3_16 = ap3Helper[4];
				Point3f p3_17 = ap3Helper[5];
				
				
				BezierTriangle bt = new BezierTriangle(ap3CoonsPatch[0],ap3CoonsPatch[1],ap3CoonsPatch[2],ap3CoonsPatch[3],ap3CoonsPatch[4],ap3CoonsPatch[5],ap3CoonsPatch[6],ap3CoonsPatch[7],ap3CoonsPatch[8]);
				Point3f p3_18 = bt.getCenterPoint();
				/*
				Vector3f v3_n18 = bt.getCenterNormal();
				v3_n18.normalize();
				Plane plane = new Plane(p3_18,v3_n18);
				*/
				
				/*
				Point3f p3_18 = new Point3f(p3_3);
				p3_18.add(p3_9);
				p3_18.add(p3_15);
				p3_18.scale(1f/3f);
				*/
				/*
				Plane plane = new Plane(p3_18,Plane.normal(p3_3,p3_9,p3_15));
				
				v3_a = Functions.vector(p3_0,p3_17);
				v3_b = Functions.vector(p3_6,p3_7);
				v3_3.interpolate(v3_a,v3_b,0.5f);
				v3_3.scale(0.5f);
				Point3f p3_24 = new Point3f(p3_3);
				p3_24.add(v3_3);
				Point3f p3_23 = new Point3f(p3_18);
				p3_23.sub(v3_3);
				Point3f p3_27 = new Point3f(p3_18);
				p3_27.add(v3_3);
				p3_27.add(v3_3);
				
				v3_a = Functions.vector(p3_6,p3_5);
				v3_b = Functions.vector(p3_12,p3_13);
				v3_9.interpolate(v3_a,v3_b,0.5f);
				v3_9.scale(0.5f);
				Point3f p3_20 = new Point3f(p3_9);
				p3_20.add(v3_9);
				Point3f p3_19 = new Point3f(p3_18);
				p3_19.sub(v3_9);
				Point3f p3_25 = new Point3f(p3_18);
				p3_25.add(v3_9);
				p3_25.add(v3_9);
				
				v3_a = Functions.vector(p3_0,p3_1);
				v3_b = Functions.vector(p3_12,p3_11);
				v3_15.interpolate(v3_a,v3_b,0.5f);
				v3_15.scale(0.5f);
				Point3f p3_22 = new Point3f(p3_15);
				p3_22.add(v3_15);
				Point3f p3_21 = new Point3f(p3_18);
				p3_21.sub(v3_15);
				Point3f p3_26 = new Point3f(p3_18);
				p3_26.add(v3_15);
				p3_26.add(v3_15);
				
				p3_25 = plane.projectedPoint(p3_25);
				p3_26 = plane.projectedPoint(p3_26);
				p3_27 = plane.projectedPoint(p3_27);
				
				p3_19 = plane.projectedPoint(p3_19);
				p3_21 = plane.projectedPoint(p3_21);
				p3_23 = plane.projectedPoint(p3_23);
				
				/*
				Line2f line1;
				Line2f line2;
				line1 = new Line2f(plane.getP2(p3_18),plane.getP2(p3_19));
				line2 = new Line2f(plane.getP2(p3_26),plane.getP2(p3_27));
				p3_19 = plane.getP3(line1.intersection(line2));
				line1 = new Line2f(plane.getP2(p3_18),plane.getP2(p3_21));
				line2 = new Line2f(plane.getP2(p3_27),plane.getP2(p3_25));
				p3_21 = plane.getP3(line1.intersection(line2));
				line1 = new Line2f(plane.getP2(p3_18),plane.getP2(p3_23));
				line2 = new Line2f(plane.getP2(p3_25),plane.getP2(p3_26));
				p3_23 = plane.getP3(line1.intersection(line2));
				*/
				
				/*
				Point3f p3_19 = new Point3f();
				//Point3f p3_20 = new Point3f();
				Point3f p3_21 = new Point3f();
				//Point3f p3_22 = new Point3f();
				Point3f p3_23 = new Point3f();
				//Point3f p3_24 = new Point3f();
				
				p3_19.interpolate(p3_18,p3_9,1f/3f);
				//p3_20.interpolate(p3_18,p3_9,2f/3f);
				p3_21.interpolate(p3_18,p3_15,1f/3f);
				//p3_22.interpolate(p3_18,p3_15,2f/3f);
				p3_23.interpolate(p3_18,p3_3,1f/3f);
				//p3_24.interpolate(p3_18,p3_3,2f/3f);
				*/
				
				/*
				Point3f p3_c1_c8 = new Point3f();
				Point3f p3_c2_c7 = new Point3f();
				Point3f p3_c2_c4 = new Point3f();
				Point3f p3_c1_c5 = new Point3f();
				Point3f p3_c5_c7 = new Point3f();
				Point3f p3_c4_c8 = new Point3f();
				
				Point3f p3_c1_ = new Point3f(ap3CoonsPatch[1]);
				Point3f p3_c2_ = new Point3f(ap3CoonsPatch[2]);
				Point3f p3_c4_ = new Point3f(ap3CoonsPatch[4]);
				Point3f p3_c5_ = new Point3f(ap3CoonsPatch[5]);
				Point3f p3_c7_ = new Point3f(ap3CoonsPatch[7]);
				Point3f p3_c8_ = new Point3f(ap3CoonsPatch[8]);
				
				Vector3f v3;
				v3 = Functions.vector(ap3CoonsPatch[0],ap3CoonsPatch[1]);
				v3.scale(2f/3f);
				p3_c8_.add(v3);
				v3.scale(1f/2f);
				p3_c7_.add(v3);
				v3 = Functions.vector(ap3CoonsPatch[3],ap3CoonsPatch[2]);
				v3.scale(2f/3f);
				p3_c4_.add(v3);
				v3.scale(1f/2f);
				p3_c5_.add(v3);
				ap3Helper = Bezier.deCasteljau(ap3CoonsPatch[7],p3_c7_,p3_c5_,ap3CoonsPatch[5],0.5f);
				p3_c5_c7 = ap3Helper[3];
				ap3Helper = Bezier.deCasteljau(ap3CoonsPatch[4],p3_c4_,p3_c8_,ap3CoonsPatch[8],0.5f);
				p3_c4_c8 = ap3Helper[3];
				
				//p3_c5_c7.interpolate(p3_c5_,p3_c7_,0.5f);
				//p3_c4_c8.interpolate(p3_c4_,p3_c8_,0.5f);
				
				p3_c1_ = new Point3f(ap3CoonsPatch[1]);
				p3_c2_ = new Point3f(ap3CoonsPatch[2]);
				p3_c4_ = new Point3f(ap3CoonsPatch[4]);
				p3_c5_ = new Point3f(ap3CoonsPatch[5]);
				p3_c7_ = new Point3f(ap3CoonsPatch[7]);
				p3_c8_ = new Point3f(ap3CoonsPatch[8]);
				
				v3 = Functions.vector(ap3CoonsPatch[6],ap3CoonsPatch[5]);
				v3.scale(2f/3f);
				p3_c7_.add(v3);
				v3.scale(1f/2f);
				p3_c8_.add(v3);
				v3 = Functions.vector(ap3CoonsPatch[3],ap3CoonsPatch[4]);
				v3.scale(2f/3f);
				p3_c2_.add(v3);
				v3.scale(1f/2f);
				p3_c1_.add(v3);
				ap3Helper = Bezier.deCasteljau(ap3CoonsPatch[1],p3_c1_,p3_c8_,ap3CoonsPatch[8],0.5f);
				p3_c1_c8 = ap3Helper[3];
				ap3Helper = Bezier.deCasteljau(ap3CoonsPatch[2],p3_c2_,p3_c7_,ap3CoonsPatch[7],0.5f);
				p3_c2_c7 = ap3Helper[3];
				//p3_c1_c8.interpolate(p3_c1_,p3_c8_,0.5f);
				//p3_c2_c7.interpolate(p3_c2_,p3_c7_,0.5f);
				
				p3_c1_ = new Point3f(ap3CoonsPatch[1]);
				p3_c2_ = new Point3f(ap3CoonsPatch[2]);
				p3_c4_ = new Point3f(ap3CoonsPatch[4]);
				p3_c5_ = new Point3f(ap3CoonsPatch[5]);
				p3_c7_ = new Point3f(ap3CoonsPatch[7]);
				p3_c8_ = new Point3f(ap3CoonsPatch[8]);
				
				v3 = Functions.vector(ap3CoonsPatch[6],ap3CoonsPatch[7]);
				v3.scale(2f/3f);
				p3_c5_.add(v3);
				v3.scale(1f/2f);
				p3_c4_.add(v3);
				v3 = Functions.vector(ap3CoonsPatch[0],ap3CoonsPatch[8]);
				v3.scale(2f/3f);
				p3_c1_.add(v3);
				v3.scale(1f/2f);
				p3_c2_.add(v3);
				ap3Helper = Bezier.deCasteljau(ap3CoonsPatch[1],p3_c1_,p3_c5_,ap3CoonsPatch[5],0.5f);
				p3_c1_c5 = ap3Helper[3];
				ap3Helper = Bezier.deCasteljau(ap3CoonsPatch[2],p3_c2_,p3_c4_,ap3CoonsPatch[4],0.5f);
				p3_c2_c4 = ap3Helper[3];
				//p3_c1_c5.interpolate(p3_c1_,p3_c5_,0.5f);
				//p3_c2_c4.interpolate(p3_c2_,p3_c4_,0.5f);
				
				/*        
				p3_c1_c8.interpolate(ap3CoonsPatch[1],ap3CoonsPatch[8],0.5f);
				p3_c2_c7.interpolate(ap3CoonsPatch[2],ap3CoonsPatch[7],0.5f);
				p3_c2_c4.interpolate(ap3CoonsPatch[2],ap3CoonsPatch[4],0.5f);
				p3_c1_c5.interpolate(ap3CoonsPatch[1],ap3CoonsPatch[5],0.5f);
				
				
				BezierTriangle bt = new BezierTriangle(ap3CoonsPatch[0],ap3CoonsPatch[1],ap3CoonsPatch[2],ap3CoonsPatch[3],ap3CoonsPatch[4],ap3CoonsPatch[5],ap3CoonsPatch[6],ap3CoonsPatch[7],ap3CoonsPatch[8]);
				System.out.println("Center: " + bt.getCenterPoint());
				
				Point3f p3_18 = new Point3f();
				ap3Helper = Bezier.deCasteljau(p3_9,p3_c2_c7,p3_c1_c8,p3_0,1f/3f);
				Point3f p3_20 = ap3Helper[1];
				Point3f p3_19 = ap3Helper[2];
				//Point3f p3_18 = ap3Helper[3];
				//ap3Helper[3].scale(1f/3f);
				p3_18.add(ap3Helper[3]);
				Point3f p3_25 = ap3Helper[4];
				Point3f p3_28 = ap3Helper[5];
				System.out.println("p 18\n-----------------");
				System.out.println(ap3Helper[3]);
				
				ap3Helper = Bezier.deCasteljau(p3_15,p3_c1_c5,p3_c2_c4,p3_6,1f/3f);
				Point3f p3_22 = ap3Helper[1];
				Point3f p3_21 = ap3Helper[2];
				Point3f p3_26 = ap3Helper[4];
				Point3f p3_29 = ap3Helper[5];
				//ap3Helper[3].scale(1f/3f);
				//p3_18.add(ap3Helper[3]);
				System.out.println(ap3Helper[3]);
				
				ap3Helper = Bezier.deCasteljau(p3_3,p3_c4_c8,p3_c5_c7,p3_12,1f/3f);
				Point3f p3_24 = ap3Helper[1];
				Point3f p3_23 = ap3Helper[2];
				Point3f p3_27 = ap3Helper[4];
				Point3f p3_30 = ap3Helper[5];
				//ap3Helper[3].scale(1f/3f);
				//p3_18.add(ap3Helper[3]);
				System.out.println(ap3Helper[3]);
				
				
				v3_a = Functions.vector(p3_0,p3_17);
				v3_b = Functions.vector(p3_6,p3_7);
				v3_3.interpolate(v3_a,v3_b,0.5f);
				v3_3.scale(0.5f);
				Point3f p3_24 = new Point3f(p3_3);
				p3_24.add(v3_3);
				
				v3_a = Functions.vector(p3_6,p3_5);
				v3_b = Functions.vector(p3_12,p3_13);
				v3_9.interpolate(v3_a,v3_b,0.5f);
				v3_9.scale(0.5f);
				Point3f p3_20 = new Point3f(p3_9);
				p3_20.add(v3_9);
				
				v3_a = Functions.vector(p3_0,p3_1);
				v3_b = Functions.vector(p3_12,p3_11);
				v3_15.interpolate(v3_a,v3_b,0.5f);
				v3_15.scale(0.5f);
				Point3f p3_22 = new Point3f(p3_15);
				p3_22.add(v3_15);
				/*
				v3_3.scale(2f);
				v3_9.scale(2f);
				v3_15.scale(2f);
				
				Point3f p3_3_ = new Point3f(p3_3);
				Point3f p3_9_ = new Point3f(p3_9);
				Point3f p3_15_ = new Point3f(p3_15);
				
				p3_3_.add(v3_3);
				p3_9_.add(v3_9);
				p3_15_.add(v3_15);
				
				p3_3_.scale(1f/3f);
				p3_9_.scale(1f/3f);
				p3_15_.scale(1f/3f);
				
				//Point3f p3_18 = new Point3f(p3_3_);
				//p3_18.add(p3_9_);
				//p3_18.add(p3_15_);
				/*
				Plane plane = new Plane(p3_18,Plane.normal(p3_0,p3_6,p3_12));
				
				v3_3.scale(4f/3f);
				v3_9.scale(4f/3f);
				v3_15.scale(4f/3f);
				
				p3_3_ = new Point3f(p3_3);
				p3_9_ = new Point3f(p3_9);
				p3_15_ = new Point3f(p3_15);
				
				p3_3_.add(v3_3);
				p3_9_.add(v3_9);
				p3_15_.add(v3_15);
				
				Point3f p3_23 = plane.projectedPoint(p3_3_);
				Point3f p3_19 = plane.projectedPoint(p3_9_);
				Point3f p3_21 = plane.projectedPoint(p3_15_);
				
				
				
				
				
				Plane plane = new Plane(p3_18,Plane.normal(p3_19,p3_21,p3_23));
				p3_19 = plane.projectedPoint(p3_19);
				p3_21 = plane.projectedPoint(p3_21);
				p3_23 = plane.projectedPoint(p3_23);
				
				p3_25 = (new Plane(p3_18,p3_21,p3_23)).projectedPoint(p3_25);
				p3_26 = (new Plane(p3_18,p3_23,p3_19)).projectedPoint(p3_26);
				p3_27 = (new Plane(p3_18,p3_19,p3_21)).projectedPoint(p3_27);
				
				Line2f line1;
				Line2f line2;
				line1 = new Line2f(plane.getP2(p3_18),plane.getP2(p3_19));
				line2 = new Line2f(plane.getP2(p3_26),plane.getP2(p3_27));
				p3_19 = plane.getP3(line1.intersection(line2));
				line1 = new Line2f(plane.getP2(p3_18),plane.getP2(p3_21));
				line2 = new Line2f(plane.getP2(p3_27),plane.getP2(p3_25));
				p3_21 = plane.getP3(line1.intersection(line2));
				line1 = new Line2f(plane.getP2(p3_18),plane.getP2(p3_23));
				line2 = new Line2f(plane.getP2(p3_25),plane.getP2(p3_26));
				p3_23 = plane.getP3(line1.intersection(line2));
				
				
				Vector3f v3_0_9 = Functions.vector(p3_0,p3_9);
				Vector3f v3_6_15 = Functions.vector(p3_6,p3_15);
				Vector3f v3_12_3 = Functions.vector(p3_12,p3_3);
				v3_0_9.normalize();
				v3_6_15.normalize();
				v3_12_3.normalize();
				
				Point3f p3_19 = new Point3f(p3_18);
				Vector3f v3_19 = new Vector3f(v3_0_9);
				v3_19.scale(p3_18.distance(p3_9));
				v3_19.scale(1f/3f);
				p3_19.add(v3_19);
				Point3f p3_25 = new Point3f(p3_18);
				Vector3f v3_25 = new Vector3f(v3_0_9);
				v3_25.scale(p3_18.distance(p3_0));
				v3_25.scale(-1f/3f);
				p3_25.add(v3_25);
				
				Point3f p3_21 = new Point3f(p3_18);
				Vector3f v3_21 = new Vector3f(v3_6_15);
				v3_21.scale(p3_18.distance(p3_15));
				v3_21.scale(1f/3f);
				p3_21.add(v3_21);
				Point3f p3_26 = new Point3f(p3_18);
				Vector3f v3_26 = new Vector3f(v3_6_15);
				v3_26.scale(p3_18.distance(p3_6));
				v3_26.scale(-1f/3f);
				p3_26.add(v3_26);
				
				Point3f p3_23 = new Point3f(p3_18);
				Vector3f v3_23 = new Vector3f(v3_12_3);
				v3_23.scale(p3_18.distance(p3_3));
				v3_23.scale(1f/3f);
				p3_23.add(v3_23);
				Point3f p3_27 = new Point3f(p3_18);
				Vector3f v3_27 = new Vector3f(v3_12_3);
				v3_27.scale(p3_18.distance(p3_12));
				v3_27.scale(-1f/3f);
				p3_27.add(v3_27);
				
				
				
				Plane plane = new Plane(p3_18,Plane.normal(p3_19,p3_21,p3_23));
				p3_19 = plane.projectedPoint(p3_19);
				p3_21 = plane.projectedPoint(p3_21);
				p3_23 = plane.projectedPoint(p3_23);
				
				Line2f line1;
				Line2f line2;
				line1 = new Line2f(plane.getP2(p3_18),plane.getP2(p3_19));
				line2 = new Line2f(plane.getP2(p3_26),plane.getP2(p3_27));
				p3_19 = plane.getP3(line1.intersection(line2));
				line1 = new Line2f(plane.getP2(p3_18),plane.getP2(p3_21));
				line2 = new Line2f(plane.getP2(p3_27),plane.getP2(p3_25));
				p3_21 = plane.getP3(line1.intersection(line2));
				line1 = new Line2f(plane.getP2(p3_18),plane.getP2(p3_23));
				line2 = new Line2f(plane.getP2(p3_25),plane.getP2(p3_26));
				p3_23 = plane.getP3(line1.intersection(line2));
				*/
				/*
				aap3Patches = new Point3f[3][16];
				
				aap3Patches[0][0] = new Point3f(p3_0);
				aap3Patches[0][1] = new Point3f(p3_1);
				aap3Patches[0][2] = new Point3f(p3_2);
				aap3Patches[0][3] = new Point3f(p3_3);
				aap3Patches[0][7] = new Point3f(p3_24);
				aap3Patches[0][11] = new Point3f(p3_23);
				aap3Patches[0][15] = new Point3f(p3_18);
				aap3Patches[0][14] = new Point3f(p3_21);
				aap3Patches[0][13] = new Point3f(p3_22);
				aap3Patches[0][12] = new Point3f(p3_15);
				aap3Patches[0][8] = new Point3f(p3_16);
				aap3Patches[0][4] = new Point3f(p3_17);
				BezierPatch.computeInnerControlPoints(aap3Patches[0]);
				
				//aap3Patches[0][5] = parallelogram(aap3Patches[0][0],aap3Patches[0][1],aap3Patches[0][4]);
				//aap3Patches[0][6] = parallelogram(aap3Patches[0][3],aap3Patches[0][2],aap3Patches[0][7]);
				//aap3Patches[0][9] = parallelogram(aap3Patches[0][12],aap3Patches[0][8],aap3Patches[0][13]);
				
				aap3Patches[1][0] = new Point3f(p3_6);
				aap3Patches[1][1] = new Point3f(p3_7);
				aap3Patches[1][2] = new Point3f(p3_8);
				aap3Patches[1][3] = new Point3f(p3_9);
				aap3Patches[1][7] = new Point3f(p3_20);
				aap3Patches[1][11] = new Point3f(p3_19);
				aap3Patches[1][15] = new Point3f(p3_18);
				aap3Patches[1][14] = new Point3f(p3_23);
				aap3Patches[1][13] = new Point3f(p3_24);
				aap3Patches[1][12] = new Point3f(p3_3);
				aap3Patches[1][8] = new Point3f(p3_4);
				aap3Patches[1][4] = new Point3f(p3_5);
				BezierPatch.computeInnerControlPoints(aap3Patches[1]);
				
				//aap3Patches[1][5] = parallelogram(aap3Patches[1][0],aap3Patches[1][1],aap3Patches[1][4]);
				//aap3Patches[1][6] = parallelogram(aap3Patches[1][3],aap3Patches[1][2],aap3Patches[1][7]);
				//aap3Patches[1][9] = parallelogram(aap3Patches[1][12],aap3Patches[1][8],aap3Patches[1][13]);
				
				aap3Patches[2][0] = new Point3f(p3_12);
				aap3Patches[2][1] = new Point3f(p3_13);
				aap3Patches[2][2] = new Point3f(p3_14);
				aap3Patches[2][3] = new Point3f(p3_15);
				aap3Patches[2][7] = new Point3f(p3_22);
				aap3Patches[2][11] = new Point3f(p3_21);
				aap3Patches[2][15] = new Point3f(p3_18);
				aap3Patches[2][14] = new Point3f(p3_19);
				aap3Patches[2][13] = new Point3f(p3_20);
				aap3Patches[2][12] = new Point3f(p3_9);
				aap3Patches[2][8] = new Point3f(p3_10);
				aap3Patches[2][4] = new Point3f(p3_11);
				BezierPatch.computeInnerControlPoints(aap3Patches[2]);
				
				//aap3Patches[2][5] = parallelogram(aap3Patches[2][0],aap3Patches[2][1],aap3Patches[2][4]);
				//aap3Patches[2][6] = parallelogram(aap3Patches[2][3],aap3Patches[2][2],aap3Patches[2][7]);
				//aap3Patches[2][9] = parallelogram(aap3Patches[2][12],aap3Patches[2][8],aap3Patches[2][13]);

				//Plane plane = new Plane(p3_18,Plane.normal(p3_19,p3_21,p3_23));
				
				aap3Patches[0][10] = new Point3f(p3_25);
				aap3Patches[1][10] = new Point3f(p3_26);
				aap3Patches[2][10] = new Point3f(p3_27);
				/*
				aap3Patches[0][5] = new Point3f(p3_28);
				aap3Patches[1][5] = new Point3f(p3_29);
				aap3Patches[2][5] = new Point3f(p3_30);
				*/
				/*
				aap3Patches[0][10] = plane.projectedPoint(p3_25);
				aap3Patches[1][10] = plane.projectedPoint(p3_26);
				aap3Patches[2][10] = plane.projectedPoint(p3_27);
				
				/*
				aap3Patches[0][5] = plane.projectedPoint(p3_28);
				aap3Patches[1][5] = plane.projectedPoint(p3_29);
				aap3Patches[2][5] = plane.projectedPoint(p3_30);
				/
				/*
				aap3Patches[0][10] = new Point3f(p3_25);
				aap3Patches[1][10] = new Point3f(p3_26);
				aap3Patches[2][10] = new Point3f(p3_27);
				*/
				/*
				aap3Patches[0][5] = new Point3f(p3_28);
				aap3Patches[1][5] = new Point3f(p3_29);
				aap3Patches[2][5] = new Point3f(p3_30);
				*/
				/*
				aap3Patches[0][7].interpolate(aap3Patches[0][6],aap3Patches[1][9],0.5f);
				aap3Patches[1][13].set(aap3Patches[0][7]);
				//aap3Patches[0][11].interpolate(aap3Patches[0][10],aap3Patches[1][10],0.5f);
				//aap3Patches[1][14].set(aap3Patches[0][11]);
				
				aap3Patches[1][7].interpolate(aap3Patches[1][6],aap3Patches[2][9],0.5f);
				aap3Patches[2][13].set(aap3Patches[1][7]);
				//aap3Patches[1][11].interpolate(aap3Patches[1][10],aap3Patches[2][10],0.5f);
				//aap3Patches[2][14].set(aap3Patches[1][11]);
				
				aap3Patches[2][7].interpolate(aap3Patches[2][6],aap3Patches[0][9],0.5f);
				aap3Patches[0][13].set(aap3Patches[2][7]);
				//aap3Patches[2][11].interpolate(aap3Patches[2][10],aap3Patches[0][10],0.5f);
				//aap3Patches[0][14].set(aap3Patches[2][11]);
				*/
				//System.out.println(p3_19 + " " + p3_20 + " " + p3_21 + " " + p3_22 + " " + p3_23 + " " + p3_24);
			break;
		}
		return aap3Patches;
	}
}
