package net.sf.jpatch.render;

import java.awt.*;
import java.awt.geom.*;
import java.util.*;
import javax.vecmath.*;

import net.sf.jpatch.auxilary.*;
import net.sf.jpatch.entity.*;

public class RBezierPatch {
	/** the virtual screen width (incl. supersampling)**/
	private static int iWidth;
	/** the virtual screen height (incl. supersampling)**/
	private static int iHeight;
	/** the virtual screen centers X position (incl. supersampling)**/
	private static float fCenterX;
	/** the virtual screen centters Y position (incl. supersampling)**/
	private static float fCenterY;
	/** the distance to the z clipping plane **/
	private static float fZClipDistance;
	/** the focal length in screen units (pixel) **/
	private static float fFocalLength;
	
	/** an array holding the 16 bezier control points **/
	private Point3f[] P;
	/** an array holding the 16 control points projected to screen space **/
	private Point2D.Float[] Pscreen;
	/** the material properties **/
	private MaterialProperties materialProperties;
	/** an array list holding the triangles (after tesselation) **/
	private ArrayList listTriangles = new ArrayList();
	
	private double dUpperBounds;
	private double dLowerBounds;
	private boolean bReady = false;
	
	/**
	* Initializes static fields
	* @param width the virtual screen width (incl. oversampling)
	* @param height the virtual screen height (incl. oversampling)
	* @param clip the distance to the z clipping plane
	* @param focalLength the focal length in screen units (pixel)
	**/
	public static final void init(int width, int height, float clip, float focalLength) {
		iWidth = width;
		iHeight = height;
		fCenterX = (float) width / 2;
		fCenterY = (float) height / 2;
		fZClipDistance = clip;
		fFocalLength = focalLength;
	}
	
	/**
	* Constructor
	* @param bezierControlPoints and array of 16 Point3f containing the bezier control points
	* @param materialProperties the materialProperties of this patch
	**/
	public RBezierPatch(Point3f[] bezierControlPoints, MaterialProperties materialProperties) {
		P = bezierControlPoints;
		this.materialProperties = materialProperties;
	}
	
	/**
	* Subdivides this patch into four smaller patches
	* @return ArrayList containing the four new patches
	**/
	public final ArrayList split() {
		Point3f[] A = deCasteljauSplit(P[0],P[1],P[2],P[3]);
		Point3f[] B = deCasteljauSplit(P[4],P[5],P[6],P[7]);
		Point3f[] C = deCasteljauSplit(P[8],P[9],P[10],P[11]);
		Point3f[] D = deCasteljauSplit(P[12],P[13],P[14],P[15]);
		
		Point3f[] E = deCasteljauSplit(A[0],B[0],C[0],D[0]);
		Point3f[] F = deCasteljauSplit(A[1],B[1],C[1],D[1]);
		Point3f[] G = deCasteljauSplit(A[2],B[2],C[2],D[2]);
		Point3f[] H = deCasteljauSplit(A[3],B[3],C[3],D[3]);
		Point3f[] I = deCasteljauSplit(A[4],B[4],C[4],D[4]);
		Point3f[] J = deCasteljauSplit(A[5],B[5],C[5],D[5]);
		Point3f[] K = deCasteljauSplit(A[6],B[6],C[6],D[6]);
		
		ArrayList patches = new ArrayList();
		patches.add(new RBezierPatch(new Point3f[] { E[0],E[1],E[2],E[3],F[0],F[1],F[2],F[3],G[0],G[1],G[2],G[3],H[0],H[1],H[2],H[3] }, materialProperties));
		patches.add(new RBezierPatch(new Point3f[] { E[3],E[4],E[5],E[6],F[3],F[4],F[5],F[6],G[3],G[4],G[5],G[6],H[3],H[4],H[5],H[6] }, materialProperties));
		patches.add(new RBezierPatch(new Point3f[] { H[0],H[1],H[2],H[3],I[0],H[1],H[2],I[3],J[0],J[1],J[2],J[3],K[0],K[1],K[2],K[3] }, materialProperties));
		patches.add(new RBezierPatch(new Point3f[] { H[3],H[4],H[5],H[6],I[3],I[4],I[5],I[6],J[3],J[4],J[5],J[6],K[3],K[4],K[5],K[6] }, materialProperties));
		
		return patches;
	}
	
	public boolean isReady() {
		return bReady;
	}
	
	public final boolean isBehindClippingPlane() {
		for (int i = 0; i < 16; i++) {
			if (P[i].z > fZClipDistance) return false;
		}
		return true;
	}
	
	/**
	* Checks if this patch needs to be split because it intersects the clipping plane
	* @return ture if the patch intersects the clipping plane
	**/
	public final boolean needsZSplit() {
		for (int i = 0; i < 16; i++) {
			if (P[i].z < fZClipDistance) return true;
		}
		return false;
	}
	
	/**
	* Performs perspective projection
	**/
	public final void project() {
		Pscreen = new Point2D.Float[16];
		for (int i = 0; i < 16; i++) {
			Pscreen[i] = new Point2D.Float(P[i].x / P[i].z * fFocalLength, P[i].y / P[i].z * fFocalLength);
		}
	}
	
	/**
	* Checks if this patch needs to be split because it intersects the vieing frustrum
	* @return true if the patch intersects the viewing frustrum
	**/
	public final boolean needsSplit() {
		float maxx = (float) iWidth * 0.5f;
		float minx = -maxx;
		float maxy = (float) iHeight * 0.5f;
		float miny = -maxy;
		
		for (int i = 0; i < 16; i++) {
			if (Pscreen[i].x < minx || Pscreen[i].x > maxx || Pscreen[i].y < miny || Pscreen[i].y > maxy) return true;
		}
		return false;
	}
	
	public final boolean isVisible() {
		float maxx = (float) iWidth * 0.5f;
		float minx = -maxx;
		float maxy = (float) iHeight * 0.5f;
		float miny = -maxy;
		for (int i = 0; i < 16; i++) {
			if (Pscreen[i].x >= minx && Pscreen[i].x <= maxx && Pscreen[i].y >= miny || Pscreen[i].y <= maxy) return true;
		}
		return false;
	}
	
	/**
	* Checks if this patch needs to be split because it is larger than desired
	* @param size desired size
	* @return true if the patch's screen-size is larger than size
	**/
	
	public final boolean needsSplit(float size) {
		return (getSizeOnScreen() > size);
	}
	
	/**
	* Computes the bounding rectangle (in screen space) of this patch
	* @return the bounding rectangle
	**/
	public final Rectangle2D.Float getBoundingRectangle() {
		float maxx = -Float.MAX_VALUE;
		float minx = Float.MAX_VALUE;
		float maxy = -Float.MAX_VALUE;
		float miny = Float.MAX_VALUE;
		//project();
		for (int i = 0; i < 16; i++) {
			if (Pscreen[i].x < minx) minx = Pscreen[i].x;
			if (Pscreen[i].x > maxx) maxx = Pscreen[i].x;
			if (Pscreen[i].y < miny) miny = Pscreen[i].y;
			if (Pscreen[i].y > maxy) maxy = Pscreen[i].y;
		}
		
		return new Rectangle2D.Float(minx,miny,maxx - minx,maxy - miny);
	}
	
	public Point2D.Float[] getProjectedPoints() {
		return Pscreen;
	}
	
	public final void prepare() {
		project();
		Rectangle2D.Float bounds = getBoundingRectangle();
		dLowerBounds = fCenterY - bounds.y;
		dUpperBounds = dLowerBounds - bounds.height;
	}
	
	public double getUpperBounds() {
		return dUpperBounds;
	}
	
	public double getLowerBounds() {
		return dLowerBounds;
	}
	
	public void fillScanline(RScanline scanline, int y) {
		ArrayList trianglesToRemove = new ArrayList();
		for (Iterator it = listTriangles.iterator(); it.hasNext(); ) {
			RTriangle triangle = (RTriangle) it.next();
			triangle.fillScanline(scanline,y,materialProperties);
			if (y > triangle.getLowerBounds()) {
				trianglesToRemove.add(triangle);
			}
		}
		listTriangles.removeAll(trianglesToRemove);
	}
	
	public ArrayList getTriangleList() {
		return listTriangles;
	}
	
	/**
	* Computes the estimated size of the patch on screen in screen units (pixel)
	* @return estimated size
	**/
	public final float getSizeOnScreen() {
		Rectangle2D.Float bounds = getBoundingRectangle();
		return (bounds.width > bounds.height) ? bounds.width : bounds.height;
	}
	
	/**
	* De Casteljau splits a cubic curve at alpha = 0.5
	* @param p0 1st control point
	* @param p1 2nd control point
	* @param p2 3rd control point
	* @param p3 4th control point
	* @return an array of 7 Point3f
	**/
	private static final Point3f[] deCasteljauSplit(Point3f p0, Point3f p1, Point3f p2, Point3f p3) {
		Point3f[] p = new Point3f[7];
		p[0] = new Point3f(p0);
		p[1] = new Point3f((p0.x + p1.x) * 0.5f,(p0.y + p1.y) * 0.5f,(p0.z + p1.z) * 0.5f);
		p[2] = new Point3f(p0.x * 0.25f + p1.x * 0.5f + p2.x * 0.25f,p0.y * 0.25f + p1.y * 0.5f + p2.y * 0.25f,p0.z * 0.25f + p1.z * 0.5f + p2.z * 0.25f);
		p[3] = new Point3f((p0.x + p1.x * 3 + p2.x * 3 + p3.x) * 0.125f,(p0.y + p1.y * 3 + p2.y * 3 + p3.y) * 0.125f,(p0.z + p1.z * 3 + p2.z * 3 + p3.z) * 0.125f);
		p[4] = new Point3f(p1.x * 0.25f + p2.x * 0.5f + p3.x * 0.25f,p1.y * 0.25f + p2.y * 0.5f + p3.y * 0.25f,p1.z * 0.25f + p2.z * 0.5f + p3.z * 0.25f);
		p[5] = new Point3f((p2.x + p3.x) * 0.5f,(p2.y + p3.y) * 0.5f,(p2.z + p3.z) * 0.5f);
		p[6] = new Point3f(p3);
		return p;
	}
	
	/**
	* Tesseates the patch into triangles
	* @param subdivisions the number of subdivisions 2 = 2 triangles, 3 = 8 triangles, 4 = 18 triangles,... ,9 = 128 triangles,...
	**/
	public final void tesselate(int subdivisions) {
		Tesselator.tesselate(P,subdivisions);
		Point3d[][] p3P = Tesselator.getPointArray();
		Vector3d[][] v3N = Tesselator.getNormalArray();
		for (int u = 0; u < subdivisions - 1; u++) {
			for (int v = 0; v < subdivisions - 1; v++) {
				addTriangle(p3P[u][v], p3P[u + 1][v], p3P[u][v + 1],v3N[u][v], v3N[u + 1][v], v3N[u][v + 1]);
				addTriangle(p3P[u + 1][v], p3P[u + 1][v + 1], p3P[u][v + 1],v3N[u + 1][v], v3N[u + 1][v + 1], v3N[u][v + 1]);
			}
		}
		boolean bReady = true;
	}
	
	/**
	* Splits triangle ar clipping plane if necessary and adds the resulting triangle(s) to the list
	* @param pa 1st corner
	* @param pb 2nd corner
	* @param pc 3rd corner
	* @param na 1st normal
	* @param nb 2nd normal
	* @param nc 3rd normal
	**/
	private final void splitAddTriangle(Point3d pa, Point3d pb, Point3d pc, Vector3d na, Vector3d nb, Vector3d nc) {
		System.out.println("splitAddTriangle");
		Point3d pd = new Point3d();
		Point3d pe = new Point3d();
		Vector3d nd = new Vector3d();
		Vector3d ne = new Vector3d();
		double f1;
		double f2;
		if (pa.z < fZClipDistance) {
			/* a is behind clipping plane */
			if (pb.z < fZClipDistance) {
				if (pc.z < fZClipDistance) return;	// triangle is behind us -> drop it
				/* only c is infront of clipping plane */
				f1 = (pc.z - fZClipDistance) / (pc.z - pa.z);
				f2 = (pc.z - fZClipDistance) / (pc.z - pb.z);
				pd.interpolate(pc,pa,f1);
				nd.interpolate(nc,na,f1);
				pe.interpolate(pc,pb,f2);
				ne.interpolate(nc,nb,f2);
				addTriangle(pd,pc,pe,nd,nc,ne);
			} else {
				if (pc.z < fZClipDistance) {
					/* only b is infront of clipping plane */
					f1 = (pb.z - fZClipDistance) / (pb.z - pa.z);
					f2 = (pb.z - fZClipDistance) / (pb.z - pc.z);
					pd.interpolate(pb,pa,f1);
					nd.interpolate(nb,na,f1);
					pe.interpolate(pb,nc,f2);
					ne.interpolate(nb,nc,f2);
					addTriangle(pd,pb,pe,nd,nb,ne);
				} else {
					/* b and c are infront of clipping plane */
					f1 = (pb.z - fZClipDistance) / (pb.z - pa.z);
					f2 = (pc.z - fZClipDistance) / (pc.z - pa.z);
					pd.interpolate(pb,pa,f1);
					nd.interpolate(nb,na,f1);
					pe.interpolate(pc,pa,f2);
					ne.interpolate(nc,na,f2);
					addTriangle(pd,pb,pc,nd,nb,nc);
					addTriangle(pd,pc,pe,nd,nc,ne);
				}
			}
		} else {
			/* a is infront of clipping plane */
			if (pb.z < fZClipDistance) {
				if (pc.z < fZClipDistance) {
					/* only a is infront of clipping plane */
					f1 = (pa.z - fZClipDistance) / (pa.z - pb.z);
					f2 = (pa.z - fZClipDistance) / (pa.z - pc.z);
					pd.interpolate(pa,pb,f1);
					nd.interpolate(na,nb,f1);
					pe.interpolate(pa,pc,f2);
					ne.interpolate(na,nc,f2);
					addTriangle(pa,pc,pd,na,nc,nd);
				} else {
					/* a and c are infront of clipping plane */
					f1 = (pa.z - fZClipDistance) / (pa.z - pb.z);
					f2 = (pc.z - fZClipDistance) / (pc.z - pb.z);
					pd.interpolate(pa,pb,f1);
					nd.interpolate(na,nb,f1);
					pe.interpolate(pc,nb,f2);
					ne.interpolate(nc,nb,f2);
					addTriangle(pa,pc,pd,na,nc,nd);
					addTriangle(pe,pd,pc,ne,nd,nc);
				}
			} else {
				if (pc.z < fZClipDistance) {
					/* a and b are infront of clipping plane */
					f1 = (pa.z - fZClipDistance) / (pa.z - pc.z);
					f2 = (pb.z - fZClipDistance) / (pb.z - pc.z);
					pd.interpolate(pa,pc,f1);
					nd.interpolate(na,nc,f1);
					pe.interpolate(pb,nc,f2);
					ne.interpolate(nb,nc,f2);
					addTriangle(pd,pa,pb,nd,na,nb);
					addTriangle(pe,pd,pb,ne,nd,nb);
				} else {
					/* everything is infront of clipping plane */
					addTriangle(pa,pb,pc,na,nb,nc);
				}
			}
		}
	}

	/**
	* Adds a triangle to the list
	* @param pa 1st corner
	* @param pb 2nd corner
	* @param pc 3rd corner
	* @param na 1st normal
	* @param nb 2nd normal
	* @param nc 3rd normal
	**/
	private final void addTriangle(Point3d pa, Point3d pb, Point3d pc, Vector3d na, Vector3d nb, Vector3d nc) {
		Point3d A = new Point3d();
		Point3d B = new Point3d();
		Point3d C = new Point3d();
		Point3d D = new Point3d();
		Point3d pd = new Point3d();
		Vector3d nd = new Vector3d();
		A.x = pa.x / pa.z * fFocalLength + fCenterX;
		A.y = -pa.y / pa.z * fFocalLength + fCenterY;
		B.x = pb.x / pb.z * fFocalLength + fCenterX;
		B.y = -pb.y / pb.z * fFocalLength + fCenterY;
		C.x = pc.x / pc.z * fFocalLength + fCenterX;
		C.y = -pc.y / pc.z * fFocalLength + fCenterY;
		if (A.x >= 0 && A.x < iWidth && A.y >= 0 && A.y < iHeight || B.x >= 0 && B.x < iWidth && B.y >= 0 && B.y < iHeight || C.x >= 0 && C.x < iWidth && C.y >= 0 && C.y < iHeight) {
			if (A.y < B.y && A.y < C.y) {
				if (B.y < C.y) {
					double f = (B.y - A.y) / (C.y - A.y);
					pd.interpolate(pa,pc,f);
					nd.interpolate(na,nc,f);
					D.x = pd.x / pd.z * fFocalLength + fCenterX;
					D.y = -pd.y / pd.z * fFocalLength + fCenterY;
					if (B.x < D.x) {
						listTriangles.add(new RTriangle(A.x,A.y,B.x,B.y,D.x,D.y,pa,pb,pd,na,nb,nd));
						listTriangles.add(new RTriangle(C.x,C.y,B.x,B.y,D.x,D.y,pc,pb,pd,nc,nb,nd));
					} else {
						listTriangles.add(new RTriangle(A.x,A.y,D.x,D.y,B.x,B.y,pa,pd,pb,na,nd,nb));
						listTriangles.add(new RTriangle(C.x,C.y,D.x,D.y,B.x,B.y,pc,pd,pb,nc,nd,nb));
					}
				} else if (B.y > C.y) {
					double f = (C.y - A.y) / (B.y - A.y);
					pd.interpolate(pa,pb,f);
					nd.interpolate(na,nb,f);
					D.x = pd.x / pd.z * fFocalLength + fCenterX;
					D.y = -pd.y / pd.z * fFocalLength + fCenterY;
					if (C.x < D.x) {
						listTriangles.add(new RTriangle(A.x,A.y,C.x,C.y,D.x,D.y,pa,pc,pd,na,nc,nd));
						listTriangles.add(new RTriangle(B.x,B.y,C.x,C.y,D.x,D.y,pb,pc,pd,nb,nc,nd));
					} else {
						listTriangles.add(new RTriangle(A.x,A.y,D.x,D.y,C.x,C.y,pa,pd,pc,na,nd,nc));
						listTriangles.add(new RTriangle(B.x,B.y,D.x,D.y,C.x,C.y,pb,pd,pc,nb,nd,nc));
					}
				} else {
					if (B.x < C.x) {
						listTriangles.add(new RTriangle(A.x,A.y,B.x,B.y,C.x,C.y,pa,pb,pc,na,nb,nc));
					} else {
						listTriangles.add(new RTriangle(A.x,C.y,B.x,A.y,C.x,B.y,pa,pc,pb,na,nc,nb));
					}
				}
			} else if (B.y < A.y && B.y < C.y) {
				if (A.y < C.y) {
					double f = (A.y - B.y) / (C.y - B.y);
					pd.interpolate(pb,pc,f);
					nd.interpolate(nb,nc,f);
					D.x = pd.x / pd.z * fFocalLength + fCenterX;
					D.y = -pd.y / pd.z * fFocalLength + fCenterY;
					if (A.x < D.x) {
						listTriangles.add(new RTriangle(B.x,B.y,A.x,A.y,D.x,D.y,pb,pa,pd,nb,na,nd));
						listTriangles.add(new RTriangle(C.x,C.y,A.x,A.y,D.x,D.y,pc,pa,pd,nc,na,nd));
					} else {
						listTriangles.add(new RTriangle(B.x,B.y,D.x,D.y,A.x,A.y,pb,pd,pa,nb,nd,na));
						listTriangles.add(new RTriangle(C.x,C.y,D.x,D.y,A.x,A.y,pc,pd,pa,nc,nd,na));
					}
				} else if (A.y > C.y) {
					double f = (C.y - B.y) / (A.y - B.y);
					pd.interpolate(pb,pa,f);
					nd.interpolate(nb,na,f);
					D.x = pd.x / pd.z * fFocalLength + fCenterX;
					D.y = -pd.y / pd.z * fFocalLength + fCenterY;
					if (C.x < D.x) {
						listTriangles.add(new RTriangle(B.x,B.y,C.x,C.y,D.x,D.y,pb,pc,pd,nb,nc,nd));
						listTriangles.add(new RTriangle(A.x,A.y,C.x,C.y,D.x,D.y,pa,pc,pd,na,nc,nd));
					} else {
						listTriangles.add(new RTriangle(B.x,B.y,D.x,D.y,C.x,C.y,pb,pd,pc,nb,nd,nc));
						listTriangles.add(new RTriangle(A.x,A.y,D.x,D.y,C.x,C.y,pa,pd,pc,na,nd,nc));
					}
				} else {
					if (A.x < C.x) {
						listTriangles.add(new RTriangle(B.x,B.y,A.x,A.y,C.x,C.y,pb,pa,pc,nb,na,nc));
					} else {
						listTriangles.add(new RTriangle(B.x,B.y,C.x,C.y,A.x,A.y,pb,pc,pa,nb,nc,na));
					}
				}
			} else {
				if (A.y < B.y) {
					double f = (A.y - C.y) / (B.y - C.y);
					pd.interpolate(pc,pb,f);
					nd.interpolate(nc,nb,f);
					D.x = pd.x / pd.z * fFocalLength + fCenterX;
					D.y = -pd.y / pd.z * fFocalLength + fCenterY;
					if (A.x < D.x) {
						listTriangles.add(new RTriangle(C.x,C.y,A.x,A.y,D.x,D.y,pc,pa,pd,nc,na,nd));
						listTriangles.add(new RTriangle(B.x,B.y,A.x,A.y,D.x,D.y,pb,pa,pd,nb,na,nd));
					} else {
						listTriangles.add(new RTriangle(C.x,C.y,D.x,D.y,A.x,A.y,pc,pd,pa,nc,nd,na));
						listTriangles.add(new RTriangle(B.x,B.y,D.x,D.y,A.x,A.y,pb,pd,pa,nb,nd,na));
					}
				} else if (A.y > B.y) {
					double f = (B.y - C.y) / (A.y - C.y);
					pd.interpolate(pc,pa,f);
					nd.interpolate(nc,na,f);
					D.x = pd.x / pd.z * fFocalLength + fCenterX;
					D.y = -pd.y / pd.z * fFocalLength + fCenterY;
					if (B.x < D.x) {
						listTriangles.add(new RTriangle(C.x,C.y,B.x,B.y,D.x,D.y,pc,pb,pd,nc,nb,nd));
						listTriangles.add(new RTriangle(A.x,A.y,B.x,B.y,D.x,D.y,pa,pb,pd,na,nb,nd));
					} else {
						listTriangles.add(new RTriangle(C.x,C.y,D.x,D.y,B.x,B.y,pc,pd,pb,nc,nd,nb));
						listTriangles.add(new RTriangle(A.x,A.y,D.x,D.y,B.x,B.y,pa,pd,pb,na,nd,nb));
					}
				} else {
					if (A.x < B.x) {
						listTriangles.add(new RTriangle(C.x,C.y,A.x,A.y,B.x,B.y,pc,pa,pb,nc,na,nb));
					} else {
						listTriangles.add(new RTriangle(C.x,C.y,B.x,B.y,A.x,A.y,pc,pb,pa,nc,nb,na));
					}
				}
			}
		}
	}
}
